import collections

from dagster import PythonObjectDagsterType, usable_as_dagster_type
from dagster.core.types.dagster_type import resolve_dagster_type


def test_dagster_type_decorator():
    @usable_as_dagster_type(name=None)
    class Foo(object):
        pass

    @usable_as_dagster_type()
    class Bar(object):
        pass

    @usable_as_dagster_type
    class Baaz(object):
        pass

    assert resolve_dagster_type(Foo).name == 'Foo'
    assert resolve_dagster_type(Bar).name == 'Bar'
    assert resolve_dagster_type(Baaz).name == 'Baaz'


def test_dagster_type_decorator_name_desc():
    @usable_as_dagster_type(name='DifferentName', description='desc')
    class Something(object):
        pass

    dagster_type = resolve_dagster_type(Something)
    assert dagster_type.name == 'DifferentName'
    assert dagster_type.description == 'desc'


def test_make_dagster_type():
    SomeNamedTuple = collections.namedtuple('SomeNamedTuple', 'prop')
    DagsterSomeNamedTuple = PythonObjectDagsterType(SomeNamedTuple)
    dagster_type = resolve_dagster_type(DagsterSomeNamedTuple)
    assert dagster_type.name == 'SomeNamedTuple'
    assert SomeNamedTuple(prop='foo').prop == 'foo'

    DagsterNewNameNamedTuple = PythonObjectDagsterType(SomeNamedTuple, name='OverwriteName')
    dagster_type = resolve_dagster_type(DagsterNewNameNamedTuple)
    assert dagster_type.name == 'OverwriteName'



from collections import namedtuple

import pkg_resources
import six
import yaml

from dagster import check
from dagster.core.definitions.utils import config_from_files, config_from_yaml_strings
from dagster.core.errors import DagsterInvariantViolationError
from dagster.seven import FileNotFoundError, ModuleNotFoundError  # pylint:disable=redefined-builtin
from dagster.utils import merge_dicts
from dagster.utils.backcompat import canonicalize_backcompat_args

from .mode import DEFAULT_MODE_NAME


class PresetDefinition(
    namedtuple('_PresetDefinition', 'name run_config solid_selection mode tags')
):
    '''Defines a preset configuration in which a pipeline can execute.

    Presets can be used in Dagit to load predefined configurations into the tool.

    Presets may also be used from the Python API (in a script, or in test) as follows:

    .. code-block:: python

        execute_pipeline(pipeline_def, preset='example_preset')

    Presets may also be used with the command line tools:

    .. code-block:: shell

        $ dagster pipeline execute example_pipeline --preset example_preset

    Args:
        name (str): The name of this preset. Must be unique in the presets defined on a given
            pipeline.
        run_config (Optional[dict]): A dict representing the config to set with the preset.
            This is equivalent to the ``run_config`` argument to :py:func:`execute_pipeline`.
        solid_selection (Optional[List[str]]): A list of solid subselection (including single
            solid names) to execute with the preset. e.g. ``['*some_solid+', 'other_solid']``
        mode (Optional[str]): The mode to apply when executing this preset. (default: 'default')
        tags (Optional[Dict[str, Any]]): The tags to apply when executing this preset.
    '''

    def __new__(
        cls, name, run_config=None, solid_selection=None, mode=None, tags=None,
    ):

        return super(PresetDefinition, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            run_config=run_config,
            solid_selection=check.opt_nullable_list_param(
                solid_selection, 'solid_selection', of_type=str
            ),
            mode=check.opt_str_param(mode, 'mode', DEFAULT_MODE_NAME),
            tags=check.opt_dict_param(tags, 'tags', key_type=str),
        )

    @staticmethod
    def from_files(
        name, environment_files=None, config_files=None, solid_selection=None, mode=None, tags=None
    ):
        '''Static constructor for presets from YAML files.

        Args:
            name (str): The name of this preset. Must be unique in the presets defined on a given
                pipeline.
            config_files (Optional[List[str]]): List of paths or glob patterns for yaml files
                to load and parse as the environment config for this preset.
            solid_selection (Optional[List[str]]): A list of solid subselection (including single
                solid names) to execute with the preset. e.g. ``['*some_solid+', 'other_solid']``
            mode (Optional[str]): The mode to apply when executing this preset. (default:
                'default')
            tags (Optional[Dict[str, Any]]): The tags to apply when executing this preset.

        Returns:
            PresetDefinition: A PresetDefinition constructed from the provided YAML files.

        Raises:
            DagsterInvariantViolationError: When one of the YAML files is invalid and has a parse
                error.
        '''
        check.str_param(name, 'name')
        config_files = canonicalize_backcompat_args(
            config_files, 'config_files', environment_files, 'environment_files', '0.9.0'
        )
        config_files = check.opt_list_param(config_files, 'config_files')
        solid_selection = check.opt_nullable_list_param(
            solid_selection, 'solid_selection', of_type=str
        )
        mode = check.opt_str_param(mode, 'mode', DEFAULT_MODE_NAME)

        merged = config_from_files(config_files)

        return PresetDefinition(name, merged, solid_selection, mode, tags)

    @staticmethod
    def from_yaml_strings(name, yaml_strings=None, solid_selection=None, mode=None, tags=None):
        '''Static constructor for presets from YAML strings.

        Args:
            name (str): The name of this preset. Must be unique in the presets defined on a given
                pipeline.
            yaml_strings (Optional[List[str]]): List of yaml strings to parse as the environment
                config for this preset.
            solid_selection (Optional[List[str]]): A list of solid subselection (including single
                solid names) to execute with the preset. e.g. ``['*some_solid+', 'other_solid']``
            mode (Optional[str]): The mode to apply when executing this preset. (default:
                'default')
            tags (Optional[Dict[str, Any]]): The tags to apply when executing this preset.

        Returns:
            PresetDefinition: A PresetDefinition constructed from the provided YAML strings

        Raises:
            DagsterInvariantViolationError: When one of the YAML documents is invalid and has a
                parse error.
        '''
        check.str_param(name, 'name')
        yaml_strings = check.opt_list_param(yaml_strings, 'yaml_strings', of_type=str)
        solid_selection = check.opt_nullable_list_param(
            solid_selection, 'solid_selection', of_type=str
        )
        mode = check.opt_str_param(mode, 'mode', DEFAULT_MODE_NAME)

        merged = config_from_yaml_strings(yaml_strings)

        return PresetDefinition(name, merged, solid_selection, mode, tags)

    @staticmethod
    def from_pkg_resources(
        name, pkg_resource_defs=None, solid_selection=None, mode=None, tags=None
    ):
        '''Load a preset from a package resource, using :py:func:`pkg_resources.resource_string`.

        Example:

        .. code-block:: python

            PresetDefinition.from_pkg_resources(
                name='local',
                mode='local',
                pkg_resource_defs=[
                    ('dagster_examples.airline_demo.environments', 'local_base.yaml'),
                    ('dagster_examples.airline_demo.environments', 'local_warehouse.yaml'),
                ],
            )


        Args:
            name (str): The name of this preset. Must be unique in the presets defined on a given
                pipeline.
            pkg_resource_defs (Optional[List[(str, str)]]): List of pkg_resource modules/files to
                load as environment config for this preset.
            solid_selection (Optional[List[str]]): A list of solid subselection (including single
                solid names) to execute with this partition. e.g.
                ``['*some_solid+', 'other_solid']``
            mode (Optional[str]): The mode to apply when executing this preset. (default:
                'default')
            tags (Optional[Dict[str, Any]]): The tags to apply when executing this preset.

        Returns:
            PresetDefinition: A PresetDefinition constructed from the provided YAML strings

        Raises:
            DagsterInvariantViolationError: When one of the YAML documents is invalid and has a
                parse error.
        '''
        pkg_resource_defs = check.opt_list_param(
            pkg_resource_defs, 'pkg_resource_defs', of_type=tuple
        )

        try:
            yaml_strings = [
                six.ensure_str(pkg_resources.resource_string(*pkg_resource_def))
                for pkg_resource_def in pkg_resource_defs
            ]
        except (ModuleNotFoundError, FileNotFoundError, UnicodeDecodeError) as err:
            six.raise_from(
                DagsterInvariantViolationError(
                    'Encountered error attempting to parse yaml. Loading YAMLs from '
                    'package resources {pkg_resource_defs} '
                    'on preset "{name}".'.format(pkg_resource_defs=pkg_resource_defs, name=name)
                ),
                err,
            )

        return PresetDefinition.from_yaml_strings(name, yaml_strings, solid_selection, mode, tags)

    def get_environment_yaml(self):
        '''Get the environment dict set on a preset as YAML.

        Returns:
            str: The environment dict as YAML.
        '''
        return yaml.dump(self.run_config or {}, default_flow_style=False)

    def with_additional_config(self, run_config):
        '''Return a new PresetDefinition with additional config merged in to the existing config.'''

        check.opt_nullable_dict_param(run_config, 'run_config')
        if run_config is None:
            return self
        else:
            return PresetDefinition(
                name=self.name,
                solid_selection=self.solid_selection,
                mode=self.mode,
                tags=self.tags,
                run_config=merge_dicts(self.run_config, run_config),
            )



from dagster.utils.net import is_local_uri


def test_is_local_uri():
    assert is_local_uri('localhost')
    assert is_local_uri('localhost:80')
    assert is_local_uri('localhost:8080')
    assert is_local_uri('127.0.0.1')
    assert is_local_uri('127.0.0.1:80')
    assert is_local_uri('127.0.0.1:8080')
    assert is_local_uri('0.0.0.0')
    assert is_local_uri('0.0.0.0:80')
    assert is_local_uri('0.0.0.0:8080')

    assert is_local_uri('rpc://')

    assert not is_local_uri('elementl.com')
    assert not is_local_uri('192.0.0.1')
    assert not is_local_uri('http://elementl.com:8080')
    assert not is_local_uri('tcp://elementl.com')
    assert not is_local_uri('hdfs://mycluster.internal')

    assert not is_local_uri('bad.bad')



import subprocess
import sys

import pytest


@pytest.mark.skipif(sys.version_info < (3, 6), reason='black is not available for python < 3.6')
def test_compiled_protobuf():
    expected_status = subprocess.check_output(['git', 'status', '--porcelain'])
    subprocess.check_output([sys.executable, '-m', 'dagster.grpc.compile'])
    assert expected_status == subprocess.check_output(['git', 'status', '--porcelain'])



import weakref

from dagster import check
from dagster.core.errors import DagsterLaunchFailedError
from dagster.core.host_representation import ExternalPipeline
from dagster.core.host_representation.handle import (
    GrpcServerRepositoryLocationHandle,
    ManagedGrpcPythonEnvRepositoryLocationHandle,
)
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.grpc.types import CanCancelExecutionRequest, CancelExecutionRequest, ExecuteRunArgs
from dagster.serdes import ConfigurableClass

from .base import RunLauncher

SUBPROCESS_TICK = 0.5

GRPC_REPOSITORY_LOCATION_HANDLE_TYPES = (
    GrpcServerRepositoryLocationHandle,
    ManagedGrpcPythonEnvRepositoryLocationHandle,
)


class GrpcRunLauncher(RunLauncher, ConfigurableClass):
    '''Launches runs against running GRPC servers.

    During the transition period from the previous CLI-based user process strategy to GRPC, you
    should use the :py:class`dagster.DefaultRunLauncher`, which is aware of instance- and
    repository-level settings allowing it to switch between the two strategies.
    '''

    def __init__(self, inst_data=None):
        self._instance_weakref = None
        self._inst_data = inst_data
        # Conceivably this should be a weakref.WeakValueDictionary, but this is the safest way to
        # get the semantics we want for termination
        self._run_id_to_repository_location_handle_cache = {}

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data, config_value):
        return GrpcRunLauncher(inst_data=inst_data)

    @property
    def _instance(self):
        return self._instance_weakref() if self._instance_weakref else None

    def initialize(self, instance):
        check.inst_param(instance, 'instance', DagsterInstance)
        check.invariant(self._instance is None, 'Must only call initialize once')
        # Store a weakref to avoid a circular reference / enable GC
        self._instance_weakref = weakref.ref(instance)

    def launch_run(self, instance, run, external_pipeline):
        '''Subclasses must implement this method.'''

        check.inst_param(run, 'run', PipelineRun)
        check.inst_param(external_pipeline, 'external_pipeline', ExternalPipeline)

        repository_location_handle = external_pipeline.repository_handle.repository_location_handle

        check.inst(
            repository_location_handle,
            GRPC_REPOSITORY_LOCATION_HANDLE_TYPES,
            'GrpcRunLauncher: Can\'t launch runs for pipeline not loaded from a GRPC server',
        )
        res = repository_location_handle.client.start_run(
            ExecuteRunArgs(
                pipeline_origin=external_pipeline.get_origin(),
                pipeline_run_id=run.run_id,
                instance_ref=self._instance.get_ref(),
            )
        )

        if not res.success:
            raise (
                DagsterLaunchFailedError(
                    res.message, serializable_error_info=res.serializable_error_info
                )
            )

        self._run_id_to_repository_location_handle_cache[run.run_id] = repository_location_handle
        return run

    def can_terminate(self, run_id):
        check.str_param(run_id, 'run_id')

        if run_id not in self._run_id_to_repository_location_handle_cache:
            return False

        res = self._run_id_to_repository_location_handle_cache[run_id].client.can_cancel_execution(
            CanCancelExecutionRequest(run_id=run_id)
        )

        return res.can_cancel

    def terminate(self, run_id):
        check.str_param(run_id, 'run_id')

        if run_id not in self._run_id_to_repository_location_handle_cache:
            return False

        res = self._run_id_to_repository_location_handle_cache[run_id].client.cancel_execution(
            CancelExecutionRequest(run_id=run_id)
        )

        return res.success



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['test_historical_config_type_snap 1'] = '''{
  "__class__": "ConfigTypeSnap",
  "description": "",
  "enum_values": [],
  "fields": [],
  "given_name": "kjdkfjdkfjdkj",
  "key": "ksjdkfjdkfjd",
  "kind": {
    "__enum__": "ConfigTypeKind.STRICT_SHAPE"
  },
  "scalar_kind": null,
  "type_param_keys": []
}'''



import os

import pytest


def aws_credentials_present():
    return os.getenv('AWS_ACCESS_KEY_ID') and os.getenv('AWS_SECRET_ACCESS_KEY')


aws = pytest.mark.skipif(not aws_credentials_present(), reason='Couldn\'t find AWS credentials')

nettest = pytest.mark.nettest



import pytest

from dagster.check import CheckError
from dagster.core.definitions.dependency import SolidHandle
from dagster.seven import json


def test_handle_path():
    handle = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert handle.path == ['foo', 'bar', 'baz']
    assert SolidHandle.from_path(handle.path) == handle

    handle = SolidHandle('foo', None)
    assert handle.path == ['foo']
    assert SolidHandle.from_path(handle.path) == handle


def test_handle_to_from_string():
    handle = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert handle.to_string() == 'foo.bar.baz'
    assert SolidHandle.from_string(handle.to_string()) == handle

    handle = SolidHandle('foo', None)
    assert handle.to_string() == 'foo'
    assert SolidHandle.from_string(handle.to_string()) == handle


def test_is_or_descends_from():
    ancestor = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', SolidHandle('quux', None))))
    descendant = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert not descendant.is_or_descends_from(ancestor)

    ancestor = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    descendant = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert descendant.is_or_descends_from(ancestor)

    ancestor = SolidHandle('bar', SolidHandle('foo', None))
    descendant = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert descendant.is_or_descends_from(ancestor)

    ancestor = SolidHandle('foo', None)
    descendant = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert descendant.is_or_descends_from(ancestor)

    ancestor = SolidHandle('foo', None)
    descendant = SolidHandle('foo', None)
    assert descendant.is_or_descends_from(ancestor)

    ancestor = SolidHandle('foo', None)
    descendant = SolidHandle('bar', None)
    assert not descendant.is_or_descends_from(ancestor)

    ancestor = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    descendant = SolidHandle('baz', None)
    assert not descendant.is_or_descends_from(ancestor)

    ancestor = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    descendant = SolidHandle('baz', SolidHandle('bar', None))
    assert not descendant.is_or_descends_from(ancestor)


def test_pop():
    handle = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert handle.pop(SolidHandle('foo', None)) == SolidHandle('baz', SolidHandle('bar', None))
    assert handle.pop(SolidHandle('bar', SolidHandle('foo', None))) == SolidHandle('baz', None)

    with pytest.raises(CheckError, match='does not descend from'):
        handle = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
        handle.pop(SolidHandle('quux', None))


def test_with_ancestor():
    handle = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert handle.with_ancestor(None) == handle
    assert handle.with_ancestor(SolidHandle('quux', None)) == SolidHandle(
        'baz', SolidHandle('bar', SolidHandle('foo', SolidHandle('quux', None)))
    )


def test_dict_roundtrip():
    handle = SolidHandle('baz', SolidHandle('bar', SolidHandle('foo', None)))
    assert SolidHandle.from_dict(json.loads(json.dumps(handle._asdict()))) == handle

    handle = SolidHandle('foo', None)
    assert SolidHandle.from_dict(json.loads(json.dumps(handle._asdict()))) == handle



import typing

import pytest

from dagster import (
    Any,
    Dict,
    List,
    Optional,
    PipelineDefinition,
    Set,
    String,
    Tuple,
    usable_as_dagster_type,
)
from dagster.core.errors import DagsterInvariantViolationError
from dagster.utils.temp_file import _unlink_swallow_errors
from dagster.utils.test import check_dagster_type, execute_solids_within_pipeline


def test_unlink_swallow_errors():
    _unlink_swallow_errors('32kjhb4kjsbfdkbf.jdfhks83')


def test_execute_isolated_solids_with_bad_solid_names():
    with pytest.raises(DagsterInvariantViolationError, match='but that solid was not found'):
        execute_solids_within_pipeline(PipelineDefinition([]), set(), {'foo': {'bar': 'baz'}})


def test_typing_invalid():

    for ttype in [
        typing.List,
        typing.List[str],
        typing.Optional[int],
        typing.Set,
        typing.Set[int],
        typing.Dict,
        typing.Dict[int, str],
        typing.Tuple,
        typing.Tuple[int, int],
    ]:
        with pytest.raises(DagsterInvariantViolationError):
            check_dagster_type(ttype, None)


def test_check_dagster_type():
    res = check_dagster_type(Any, 'bar')
    assert res.success

    res = check_dagster_type(Any, None)
    assert res.success

    res = check_dagster_type(str, 'bar')
    assert res.success

    res = check_dagster_type(str, 1)
    assert not res.success
    assert 'Value "1" of python type "int" must be a string.' in res.description

    res = check_dagster_type(String, 'bar')
    assert res.success

    res = check_dagster_type(list, ['foo', 'bar'])
    assert res.success

    res = check_dagster_type(List, ['foo', 'bar'])
    assert res.success

    res = check_dagster_type(List[Any], ['foo', 'bar'])
    assert res.success

    res = check_dagster_type(List[str], ['foo', 'bar'])
    assert res.success

    res = check_dagster_type(Any, {'foo': 'bar'})
    assert res.success

    res = check_dagster_type(dict, {'foo': 'bar'})
    assert res.success

    res = check_dagster_type(Dict, {'foo': 'bar'})
    assert res.success

    res = check_dagster_type(Dict[Any, Any], {'foo': 'bar'})
    assert res.success

    res = check_dagster_type(Dict[str, str], {'foo': 'bar'})
    assert res.success

    res = check_dagster_type(Dict[str, int], {'foo': 'bar'})
    assert not res.success
    assert 'Value "bar" of python type "str" must be a int.' in res.description

    res = check_dagster_type(tuple, ('foo', 'bar'))
    assert res.success

    res = check_dagster_type(Tuple, ('foo', 'bar'))
    assert res.success

    res = check_dagster_type(Tuple[Any, Any], ('foo', 'bar'))
    assert res.success

    res = check_dagster_type(Tuple[str, str], ('foo', 'bar'))
    assert res.success

    res = check_dagster_type(Tuple[str, str], ('foo', 'bar'))
    assert res.success

    res = check_dagster_type(Tuple[str, str], ('foo', 'bar', 'baz'))
    assert not res.success
    assert (
        'Tuple with key TypedPythonTupleString.String requires 2 entries, received 3 values'
    ) in res.description

    res = check_dagster_type(Set, {'foo', 'bar'})
    assert res.success

    res = check_dagster_type(Set[Any], {'foo', 'bar'})
    assert res.success

    res = check_dagster_type(Set[str], {'foo', 'bar'})
    assert res.success

    res = check_dagster_type(Set[int], {'foo'})
    assert not res.success
    assert 'Value "foo" of python type "str" must be a int.' in res.description

    res = check_dagster_type(Optional[str], 'str')
    assert res.success

    res = check_dagster_type(Optional[str], None)
    assert res.success

    @usable_as_dagster_type
    class Baz(object):
        pass

    res = check_dagster_type(Baz, Baz())
    assert res.success



from collections import namedtuple
from functools import update_wrapper

from dagster import check
from dagster.config.field_utils import check_user_facing_opt_config_param
from dagster.utils.backcompat import rename_warning


class IntermediateStorageDefinition(
    namedtuple(
        '_IntermediateStorageDefinition',
        'name is_persistent config_schema intermediate_storage_creation_fn required_resource_keys',
    )
):
    '''Defines intermediate data storage behaviors.
    Args:
        name (str): Name of the storage mode.
        is_persistent (bool): Whether the storage is persistent in a way that can cross process/node
            boundaries. Re-execution with, for example, the multiprocess executor, or with
            dagster-airflow, requires a persistent storage mode.
        config_schema (Optional[ConfigSchema]): The schema for the storage's configuration schema.
            Configuration data passed in this schema will be made available to the
            ``intermediate_storage_creation_fn`` under ``init_context.intermediate_storage_config``.
        intermediate_storage_creation_fn: (Callable[[InitIntermediateStorageContext], IntermediateStorage])
            Called to construct the storage. This function should consume the init context and emit
            a :py:class:`IntermediateStorage`.
        required_resource_keys(Optional[Set[str]]): The resources that this storage needs at runtime to function.
    '''

    def __new__(
        cls,
        name,
        is_persistent,
        required_resource_keys,
        config_schema=None,
        intermediate_storage_creation_fn=None,
    ):
        return super(IntermediateStorageDefinition, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            is_persistent=check.bool_param(is_persistent, 'is_persistent'),
            config_schema=check_user_facing_opt_config_param(config_schema, 'config_schema'),
            intermediate_storage_creation_fn=check.opt_callable_param(
                intermediate_storage_creation_fn, 'intermediate_storage_creation_fn'
            ),
            required_resource_keys=frozenset(
                check.set_param(
                    required_resource_keys if required_resource_keys else set(),
                    'required_resource_keys',
                    of_type=str,
                )
            ),
        )

    @property
    def config_field(self):
        rename_warning('config_schema', 'config_field', '0.9.0')
        return self.config_schema


def intermediate_storage(
    required_resource_keys=None, name=None, is_persistent=True, config_schema=None
):
    '''Creates an intermediate storage definition

    The decorated function will be passed as the ``intermediate_storage_creation_fn`` to a
    :py:class:`IntermediateStorageDefinition`.

    Args:
        name (str): The name of the intermediate storage.
        is_persistent (bool): Whether the storage is persistent in a way that can cross process/node
            boundaries. Re-execution with, for example, the multiprocess executor, or with
            dagster-airflow, requires a persistent storage mode.
        required_resource_keys (Optional[Set[str]]):
            The resources that this storage needs at runtime to function.
        config_schema (Optional[ConfigSchema]): The schema for the config. Configuration data available in
            `init_context.intermediate_storage_config`.
    '''

    if callable(name):
        check.invariant(is_persistent is True)
        check.invariant(config_schema is None)
        check.invariant(required_resource_keys is None)
        return _IntermediateStorageDecoratorCallable()(name)

    return _IntermediateStorageDecoratorCallable(
        name=name,
        is_persistent=is_persistent,
        config_schema=config_schema,
        required_resource_keys=required_resource_keys,
    )


class _IntermediateStorageDecoratorCallable(object):
    def __init__(
        self, name=None, is_persistent=True, config_schema=None, required_resource_keys=None
    ):
        self.name = check.opt_str_param(name, 'name')
        self.is_persistent = check.bool_param(is_persistent, 'is_persistent')
        self.config_schema = check_user_facing_opt_config_param(config_schema, 'config_schema')
        self.required_resource_keys = check.opt_set_param(
            required_resource_keys, 'required_resource_keys', of_type=str
        )

    def __call__(self, fn):
        check.callable_param(fn, 'fn')

        if not self.name:
            self.name = fn.__name__

        storage_def = IntermediateStorageDefinition(
            name=self.name,
            is_persistent=self.is_persistent,
            config_schema=self.config_schema,
            intermediate_storage_creation_fn=fn,
            required_resource_keys=self.required_resource_keys,
        )

        update_wrapper(storage_def, wrapped=fn)

        return storage_def



from dagster import Dict, InputDefinition, List, OutputDefinition, Set, Tuple, pipeline, solid
from dagster.core.snap import build_dagster_type_namespace_snapshot
from dagster.core.types.dagster_type import ALL_RUNTIME_BUILTINS, create_string_type


def test_simple_pipeline_input_dagster_type_namespace():
    SomethingType = create_string_type('SomethingType', description='desc')

    @solid(input_defs=[InputDefinition('something', SomethingType)])
    def take_something(_, something):
        return something

    @pipeline
    def simple():
        take_something()

    namespace = build_dagster_type_namespace_snapshot(simple)
    type_snap = namespace.get_dagster_type_snap('SomethingType')
    assert type_snap
    assert type_snap.key == 'SomethingType'
    assert type_snap.name == 'SomethingType'
    assert type_snap.display_name == 'SomethingType'
    assert type_snap.description == 'desc'
    assert type_snap.is_builtin is False
    assert type_snap.type_param_keys == []
    assert type_snap.loader_schema_key == SomethingType.loader_schema_key
    assert type_snap.materializer_schema_key == SomethingType.materializer_schema_key


def test_simple_pipeline_output_dagster_type_namespace():
    SomethingType = create_string_type('SomethingType')

    @solid(output_defs=[OutputDefinition(SomethingType)])
    def take_something(_):
        return 'something'

    @pipeline
    def simple():
        take_something()

    namespace = build_dagster_type_namespace_snapshot(simple)
    assert namespace.get_dagster_type_snap('SomethingType')


def test_kitchen_sink_of_collection_types_snaps():
    SomethingType = create_string_type('SomethingType')

    @solid(input_defs=[InputDefinition('somethings', List[SomethingType])])
    def take_list(_, somethings):
        return somethings

    @solid(input_defs=[InputDefinition('somethings', Set[SomethingType])])
    def take_set(_, somethings):
        return somethings

    # dict cannot be input without dep
    # see https://github.com/dagster-io/dagster/issues/2272
    @solid(output_defs=[OutputDefinition(Dict[str, SomethingType])])
    def return_dict(_):
        return {}

    @solid(input_defs=[InputDefinition('somethings', Tuple[str, SomethingType])])
    def take_tuple(_, somethings):
        return somethings

    @pipeline
    def simple():
        take_list()
        take_set()
        return_dict()
        take_tuple()

    namespace = build_dagster_type_namespace_snapshot(simple)

    assert namespace.get_dagster_type_snap(List[SomethingType].key)
    list_something = namespace.get_dagster_type_snap(List[SomethingType].key)
    assert len(list_something.type_param_keys) == 1
    assert list_something.type_param_keys[0] == SomethingType.key
    assert list_something.display_name == '[SomethingType]'

    assert namespace.get_dagster_type_snap(Set[SomethingType].key)
    something_set = namespace.get_dagster_type_snap(Set[SomethingType].key)
    assert len(something_set.type_param_keys) == 1
    assert something_set.type_param_keys[0] == SomethingType.key
    assert something_set.display_name == 'Set[SomethingType]'

    assert namespace.get_dagster_type_snap(Dict[str, SomethingType].key)
    something_dict = namespace.get_dagster_type_snap(Dict[str, SomethingType].key)
    assert len(something_dict.type_param_keys) == 2
    assert something_dict.type_param_keys[0] == 'String'
    assert something_dict.type_param_keys[1] == SomethingType.key
    assert something_dict.display_name == 'Dict[String,SomethingType]'

    assert namespace.get_dagster_type_snap(Tuple[str, SomethingType].key)
    something_tuple = namespace.get_dagster_type_snap(Tuple[str, SomethingType].key)
    assert len(something_tuple.type_param_keys) == 2
    assert something_tuple.type_param_keys[0] == 'String'
    assert something_tuple.type_param_keys[1] == SomethingType.key
    assert something_tuple.display_name == 'Tuple[String,SomethingType]'


def test_kitchen_sink_of_builtins():
    @solid
    def noop(_):
        pass

    @pipeline
    def simple():
        noop()

    namespace = build_dagster_type_namespace_snapshot(simple)

    for builtin in ALL_RUNTIME_BUILTINS:
        assert namespace.get_dagster_type_snap(builtin.key)



import os
import sys
import time
import traceback
from contextlib import contextmanager

import pytest
import sqlalchemy

from dagster import seven
from dagster.core.definitions import AssetMaterialization, ExpectationResult
from dagster.core.errors import DagsterEventLogInvalidForRun
from dagster.core.events import (
    DagsterEvent,
    DagsterEventType,
    EngineEventData,
    StepExpectationResultData,
    StepMaterializationData,
)
from dagster.core.events.log import DagsterEventRecord
from dagster.core.execution.plan.objects import StepFailureData, StepSuccessData
from dagster.core.storage.event_log import (
    ConsolidatedSqliteEventLogStorage,
    InMemoryEventLogStorage,
    SqlEventLogStorageMetadata,
    SqlEventLogStorageTable,
    SqliteEventLogStorage,
)
from dagster.core.storage.sql import create_engine
from dagster.seven import multiprocessing


@contextmanager
def create_in_memory_event_log_storage():
    yield InMemoryEventLogStorage()


@contextmanager
def create_sqlite_run_event_logstorage():
    with seven.TemporaryDirectory() as tmpdir_path:
        yield SqliteEventLogStorage(tmpdir_path)


@contextmanager
def create_consolidated_sqlite_run_event_log_storage():
    with seven.TemporaryDirectory() as tmpdir_path:
        yield ConsolidatedSqliteEventLogStorage(tmpdir_path)


event_storage_test = pytest.mark.parametrize(
    'event_storage_factory_cm_fn',
    [
        create_in_memory_event_log_storage,
        create_sqlite_run_event_logstorage,
        create_consolidated_sqlite_run_event_log_storage,
    ],
)


@event_storage_test
def test_init_log_storage(event_storage_factory_cm_fn):
    with event_storage_factory_cm_fn() as storage:
        if isinstance(storage, InMemoryEventLogStorage):
            assert not storage.is_persistent
        elif isinstance(storage, (SqliteEventLogStorage, ConsolidatedSqliteEventLogStorage)):
            assert storage.is_persistent
        else:
            raise Exception("Invalid event storage type")


@event_storage_test
def test_log_storage_run_not_found(event_storage_factory_cm_fn):
    with event_storage_factory_cm_fn() as storage:
        assert storage.get_logs_for_run('bar') == []


@event_storage_test
def test_event_log_storage_store_events_and_wipe(event_storage_factory_cm_fn):
    with event_storage_factory_cm_fn() as storage:
        assert len(storage.get_logs_for_run('foo')) == 0
        storage.store_event(
            DagsterEventRecord(
                None,
                'Message2',
                'debug',
                '',
                'foo',
                time.time(),
                dagster_event=DagsterEvent(
                    DagsterEventType.ENGINE_EVENT.value,
                    'nonce',
                    event_specific_data=EngineEventData.in_process(999),
                ),
            )
        )
        assert len(storage.get_logs_for_run('foo')) == 1
        assert storage.get_stats_for_run('foo')
        storage.wipe()
        assert len(storage.get_logs_for_run('foo')) == 0


@event_storage_test
def test_event_log_storage_store_with_multiple_runs(event_storage_factory_cm_fn):
    with event_storage_factory_cm_fn() as storage:
        runs = ['foo', 'bar', 'baz']
        for run_id in runs:
            assert len(storage.get_logs_for_run(run_id)) == 0
            storage.store_event(
                DagsterEventRecord(
                    None,
                    'Message2',
                    'debug',
                    '',
                    run_id,
                    time.time(),
                    dagster_event=DagsterEvent(
                        DagsterEventType.STEP_SUCCESS.value,
                        'nonce',
                        event_specific_data=StepSuccessData(duration_ms=100.0),
                    ),
                )
            )

        for run_id in runs:
            assert len(storage.get_logs_for_run(run_id)) == 1
            assert storage.get_stats_for_run(run_id).steps_succeeded == 1

        storage.wipe()
        for run_id in runs:
            assert len(storage.get_logs_for_run(run_id)) == 0


@event_storage_test
def test_event_log_storage_watch(event_storage_factory_cm_fn):
    def evt(name):
        return DagsterEventRecord(
            None,
            name,
            'debug',
            '',
            'foo',
            time.time(),
            dagster_event=DagsterEvent(
                DagsterEventType.ENGINE_EVENT.value,
                'nonce',
                event_specific_data=EngineEventData.in_process(999),
            ),
        )

    with event_storage_factory_cm_fn() as storage:
        watched = []
        watcher = lambda x: watched.append(x)  # pylint: disable=unnecessary-lambda

        assert len(storage.get_logs_for_run('foo')) == 0

        storage.store_event(evt('Message1'))
        assert len(storage.get_logs_for_run('foo')) == 1
        assert len(watched) == 0

        storage.watch('foo', 0, watcher)

        storage.store_event(evt('Message2'))
        storage.store_event(evt('Message3'))
        storage.store_event(evt('Message4'))

        attempts = 10
        while len(watched) < 3 and attempts > 0:
            time.sleep(0.1)
            attempts -= 1

        storage.end_watch('foo', watcher)
        time.sleep(0.3)  # this value scientifically selected from a range of attractive values
        storage.store_event(evt('Message5'))

        assert len(storage.get_logs_for_run('foo')) == 5
        assert len(watched) == 3

        storage.delete_events('foo')

        assert len(storage.get_logs_for_run('foo')) == 0
        assert len(watched) == 3


@event_storage_test
def test_event_log_storage_pagination(event_storage_factory_cm_fn):
    def evt(name):
        return DagsterEventRecord(
            None,
            name,
            'debug',
            '',
            'foo',
            time.time(),
            dagster_event=DagsterEvent(
                DagsterEventType.ENGINE_EVENT.value,
                'nonce',
                event_specific_data=EngineEventData.in_process(999),
            ),
        )

    with event_storage_factory_cm_fn() as storage:
        storage.store_event(evt('Message_0'))
        storage.store_event(evt('Message_1'))
        storage.store_event(evt('Message_2'))

        assert len(storage.get_logs_for_run('foo')) == 3
        assert len(storage.get_logs_for_run('foo', -1)) == 3
        assert len(storage.get_logs_for_run('foo', 0)) == 2
        assert len(storage.get_logs_for_run('foo', 1)) == 1
        assert len(storage.get_logs_for_run('foo', 2)) == 0


@event_storage_test
def test_event_log_delete(event_storage_factory_cm_fn):
    with event_storage_factory_cm_fn() as storage:
        assert len(storage.get_logs_for_run('foo')) == 0
        storage.store_event(
            DagsterEventRecord(
                None,
                'Message2',
                'debug',
                '',
                'foo',
                time.time(),
                dagster_event=DagsterEvent(
                    DagsterEventType.ENGINE_EVENT.value,
                    'nonce',
                    event_specific_data=EngineEventData.in_process(999),
                ),
            )
        )
        assert len(storage.get_logs_for_run('foo')) == 1
        assert storage.get_stats_for_run('foo')
        storage.delete_events('foo')
        assert len(storage.get_logs_for_run('foo')) == 0


@event_storage_test
def test_event_log_get_stats_without_start_and_success(event_storage_factory_cm_fn):
    # When an event log doesn't have a PIPELINE_START or PIPELINE_SUCCESS | PIPELINE_FAILURE event,
    # we want to ensure storage.get_stats_for_run(...) doesn't throw an error.

    with event_storage_factory_cm_fn() as storage:
        assert len(storage.get_logs_for_run('foo')) == 0
        assert storage.get_stats_for_run('foo')


def test_filesystem_event_log_storage_run_corrupted():
    with seven.TemporaryDirectory() as tmpdir_path:
        storage = SqliteEventLogStorage(tmpdir_path)
        # URL begins sqlite:///
        # pylint: disable=protected-access
        with open(os.path.abspath(storage.conn_string_for_run_id('foo')[10:]), 'w') as fd:
            fd.write('some nonsense')
        with pytest.raises(sqlalchemy.exc.DatabaseError):
            storage.get_logs_for_run('foo')


def test_filesystem_event_log_storage_run_corrupted_bad_data():
    with seven.TemporaryDirectory() as tmpdir_path:
        storage = SqliteEventLogStorage(tmpdir_path)
        SqlEventLogStorageMetadata.create_all(create_engine(storage.conn_string_for_run_id('foo')))
        with storage.connect('foo') as conn:
            event_insert = SqlEventLogStorageTable.insert().values(  # pylint: disable=no-value-for-parameter
                run_id='foo', event='{bar}', dagster_event_type=None, timestamp=None
            )
            conn.execute(event_insert)

        with pytest.raises(DagsterEventLogInvalidForRun):
            storage.get_logs_for_run('foo')

        SqlEventLogStorageMetadata.create_all(create_engine(storage.conn_string_for_run_id('bar')))

        with storage.connect('bar') as conn:  # pylint: disable=protected-access
            event_insert = SqlEventLogStorageTable.insert().values(  # pylint: disable=no-value-for-parameter
                run_id='bar', event='3', dagster_event_type=None, timestamp=None
            )
            conn.execute(event_insert)
        with pytest.raises(DagsterEventLogInvalidForRun):
            storage.get_logs_for_run('bar')


def cmd(exceptions, tmpdir_path):
    storage = SqliteEventLogStorage(tmpdir_path)
    try:
        with storage.connect('foo'):
            pass
    except Exception as exc:  # pylint: disable=broad-except
        exceptions.put(exc)
        exc_info = sys.exc_info()
        traceback.print_tb(exc_info[2])


def test_concurrent_sqlite_event_log_connections():
    exceptions = multiprocessing.Queue()

    with seven.TemporaryDirectory() as tmpdir_path:
        ps = []
        for _ in range(5):
            ps.append(multiprocessing.Process(target=cmd, args=(exceptions, tmpdir_path)))
        for p in ps:
            p.start()

        j = 0
        for p in ps:
            p.join()
            j += 1

        assert j == 5

        excs = []
        while not exceptions.empty():
            excs.append(exceptions.get())
        assert not excs, excs


@event_storage_test
def test_event_log_step_stats(event_storage_factory_cm_fn):
    # When an event log doesn't have a PIPELINE_START or PIPELINE_SUCCESS | PIPELINE_FAILURE event,
    # we want to ensure storage.get_stats_for_run(...) doesn't throw an error.

    run_id = 'foo'
    with event_storage_factory_cm_fn() as storage:
        for record in _stats_records(run_id=run_id):
            storage.store_event(record)

        step_stats = storage.get_step_stats_for_run(run_id)
        assert len(step_stats) == 4

        a_stats = [stats for stats in step_stats if stats.step_key == 'A'][0]
        assert a_stats.step_key == 'A'
        assert a_stats.status.value == 'SUCCESS'
        assert a_stats.end_time - a_stats.start_time == 100

        b_stats = [stats for stats in step_stats if stats.step_key == 'B'][0]
        assert b_stats.step_key == 'B'
        assert b_stats.status.value == 'FAILURE'
        assert b_stats.end_time - b_stats.start_time == 50

        c_stats = [stats for stats in step_stats if stats.step_key == 'C'][0]
        assert c_stats.step_key == 'C'
        assert c_stats.status.value == 'SKIPPED'
        assert c_stats.end_time - c_stats.start_time == 25

        d_stats = [stats for stats in step_stats if stats.step_key == 'D'][0]
        assert d_stats.step_key == 'D'
        assert d_stats.status.value == 'SUCCESS'
        assert d_stats.end_time - d_stats.start_time == 150
        assert len(d_stats.materializations) == 3
        assert len(d_stats.expectation_results) == 2


def _stats_records(run_id):
    now = time.time()
    return [
        _event_record(run_id, 'A', now - 325, DagsterEventType.STEP_START),
        _event_record(
            run_id,
            'A',
            now - 225,
            DagsterEventType.STEP_SUCCESS,
            StepSuccessData(duration_ms=100000.0),
        ),
        _event_record(run_id, 'B', now - 225, DagsterEventType.STEP_START),
        _event_record(
            run_id,
            'B',
            now - 175,
            DagsterEventType.STEP_FAILURE,
            StepFailureData(error=None, user_failure_data=None),
        ),
        _event_record(run_id, 'C', now - 175, DagsterEventType.STEP_START),
        _event_record(run_id, 'C', now - 150, DagsterEventType.STEP_SKIPPED),
        _event_record(run_id, 'D', now - 150, DagsterEventType.STEP_START),
        _event_record(
            run_id,
            'D',
            now - 125,
            DagsterEventType.STEP_MATERIALIZATION,
            StepMaterializationData(AssetMaterialization(asset_key='mat_1')),
        ),
        _event_record(
            run_id,
            'D',
            now - 100,
            DagsterEventType.STEP_EXPECTATION_RESULT,
            StepExpectationResultData(ExpectationResult(success=True, label='exp 1')),
        ),
        _event_record(
            run_id,
            'D',
            now - 75,
            DagsterEventType.STEP_MATERIALIZATION,
            StepMaterializationData(AssetMaterialization(asset_key='mat_2')),
        ),
        _event_record(
            run_id,
            'D',
            now - 50,
            DagsterEventType.STEP_EXPECTATION_RESULT,
            StepExpectationResultData(ExpectationResult(success=False, label='exp 2')),
        ),
        _event_record(
            run_id,
            'D',
            now - 25,
            DagsterEventType.STEP_MATERIALIZATION,
            StepMaterializationData(AssetMaterialization(asset_key='mat_3')),
        ),
        _event_record(
            run_id, 'D', now, DagsterEventType.STEP_SUCCESS, StepSuccessData(duration_ms=150000.0)
        ),
    ]


def _event_record(run_id, step_key, timestamp, event_type, event_specific_data=None):
    pipeline_name = 'pipeline_name'
    return DagsterEventRecord(
        None,
        '',
        'debug',
        '',
        run_id,
        timestamp,
        step_key=step_key,
        pipeline_name=pipeline_name,
        dagster_event=DagsterEvent(
            event_type.value,
            pipeline_name,
            step_key=step_key,
            event_specific_data=event_specific_data,
        ),
    )



import logging

from dagster import (
    Enum,
    EnumValue,
    Field,
    Int,
    ModeDefinition,
    configured,
    execute_pipeline,
    logger,
    pipeline,
)
from dagster.core.log_manager import coerce_valid_log_level


def assert_pipeline_runs_with_logger(logger_def, logger_config):
    @pipeline(mode_defs=[ModeDefinition(logger_defs={'test_logger': logger_def})])
    def pass_pipeline():
        pass

    # give us an opportunity to try an empty dict here
    config_value = {'config': logger_config} if logger_config else {}

    result = execute_pipeline(pass_pipeline, {'loggers': {'test_logger': config_value}})
    assert result.success


def test_dagster_type_logger_decorator_config():
    @logger(Int)
    def dagster_type_logger_config(_):
        raise Exception('not called')

    assert dagster_type_logger_config.config_schema.config_type.given_name == 'Int'

    @logger(int)
    def python_type_logger_config(_):
        raise Exception('not called')

    assert python_type_logger_config.config_schema.config_type.given_name == 'Int'


def test_logger_using_configured():
    it = {'ran': False}

    @logger(config_schema=Field(str))
    def test_logger(init_context):
        assert init_context.logger_config == 'secret testing value!!'
        it['ran'] = True

        logger_ = logging.Logger('test', level=coerce_valid_log_level('INFO'))
        return logger_

    test_logger_configured = configured(test_logger)('secret testing value!!')

    assert_pipeline_runs_with_logger(test_logger_configured, {})
    assert it['ran']


def test_logger_with_enum_in_schema_using_configured():
    from enum import Enum as PythonEnum

    class TestPythonEnum(PythonEnum):
        VALUE_ONE = 0
        OTHER = 1

    DagsterEnumType = Enum(
        'TestEnum',
        [
            EnumValue('VALUE_ONE', TestPythonEnum.VALUE_ONE),
            EnumValue('OTHER', TestPythonEnum.OTHER),
        ],
    )

    it = {}

    @logger(config_schema={'enum': DagsterEnumType})
    def test_logger(init_context):
        assert init_context.logger_config['enum'] == TestPythonEnum.OTHER
        it['ran test_logger'] = True

        logger_ = logging.Logger('test', level=coerce_valid_log_level('INFO'))
        return logger_

    @configured(test_logger, {'enum': DagsterEnumType})
    def pick_different_enum_value(config):
        it['ran pick_different_enum_value'] = True
        return {'enum': 'OTHER' if config['enum'] == TestPythonEnum.VALUE_ONE else 'VALUE_ONE'}

    assert_pipeline_runs_with_logger(pick_different_enum_value, {'enum': 'VALUE_ONE'})

    assert it['ran test_logger']
    assert it['ran pick_different_enum_value']



def test_ping(docker_grpc_client):
    assert docker_grpc_client.ping('foobar') == 'foobar'


def test_streaming(docker_grpc_client):
    results = [
        result for result in docker_grpc_client.streaming_ping(sequence_length=10, echo='foo')
    ]
    assert len(results) == 10
    for sequence_number, result in enumerate(results):
        assert result['sequence_number'] == sequence_number
        assert result['echo'] == 'foo'



class ConfigSchema(object):
    '''
    ConfigSchema is a placeholder type.  Any time that it appears in documentation, it means
    that any of the following types are acceptable:

    #. A Python scalar type that resolves to a Dagster config type
       (:py:class:`~python:int`, :py:class:`~python:float`, :py:class:`~python:bool`,
       or :py:class:`~python:str`

       e.g. `@solid(config_schema=int)` or `@solid(config_schema=str)`

    #. Built-in python collections such as :py:class:`~python:list`, or :py:class:`~python:dict`).
       `list` is exactly equivalent to `Array[Any]` and `dict` is equivalent to `Permissive()`

       e.g. `@solid(config_schema=list)` or `@solid(config_schema=dict)`

    #. A Dagster config type: :py:data:`~dagster.Int`, :py:data:`~dagster.IntSource`,
       :py:data:`~dagster.Float`, :py:data:`~dagster.Bool`, :py:data:`~dagster.String`,
       :py:data:`~dagster.StringSource`, :py:data:`~dagster.Any`, :py:class:`~dagster.Array`,
       :py:data:`~dagster.Noneable`, :py:data:`~dagster.Enum`, :py:class:`~dagster.Selector`,
       :py:class:`~dagster.Shape`, or :py:class:`~dagster.Permissive`.

    #. A bare python dictionary, which will be automatically wrapped in
       :py:class:`~dagster.Shape`. Values of the dictionary are resolved recursively
       according to the same rules. E.g.

        * ``{'some_config': str}`` is equivalent to ``Shape({'some_config: str})``.

        * ``{'some_config1': {'some_config2': str}}`` is equivalent to
          ``Shape({'some_config1: Shape({'some_config2: str})})``.

    #. A bare python list of length one, whose single element will be wrapped in a
       :py:class:`~dagster.Array` is resolved recursively according to the same
       rules. E.g.

        * ``[str]`` is equivalent to ``Array[str]``.

        * ``[[str]]`` is equivalent to ``Array[Array[str]]``.

        * ``[{'some_config': str}]`` is equivalent to ``Array(Shape({'some_config: str}))``.

    #. An instance of :py:class:`~dagster.Field`.
    '''

    def __init__(self):
        raise NotImplementedError(
            'ConfigSchema is a placeholder type and should not be instantiated.'
        )



from dagster import check
from dagster.core.host_representation import RepositoryLocationHandle


class Workspace:
    def __init__(self, repository_location_handles):
        check.list_param(
            repository_location_handles,
            'repository_location_handles',
            of_type=RepositoryLocationHandle,
        )
        self._location_handle_dict = {rlh.location_name: rlh for rlh in repository_location_handles}

    @property
    def repository_location_handles(self):
        return list(self._location_handle_dict.values())

    @property
    def repository_location_names(self):
        return list(self._location_handle_dict.keys())

    def has_repository_location_handle(self, location_name):
        check.str_param(location_name, 'location_name')
        return location_name in self._location_handle_dict

    def get_repository_location_handle(self, location_name):
        check.str_param(location_name, 'location_name')
        return self._location_handle_dict[location_name]



import threading
import time

from dagster.grpc.client import ephemeral_grpc_api_client


def _stream_events_target(results, api_client):
    for result in api_client.streaming_ping(sequence_length=100000, echo='foo'):
        results.append(result)


def test_streaming_terminate():
    with ephemeral_grpc_api_client() as api_client:
        streaming_results = []
        stream_events_result_thread = threading.Thread(
            target=_stream_events_target, args=[streaming_results, api_client]
        )
        stream_events_result_thread.daemon = True
        stream_events_result_thread.start()
        while not streaming_results:
            time.sleep(0.001)
        res = api_client.shutdown_server()
        assert res.success
        assert res.serializable_error_info is None

        stream_events_result_thread.join()
        assert len(streaming_results) == 100000

        api_client._server_process.wait()  # pylint: disable=protected-access
        assert api_client._server_process.poll() == 0  # pylint: disable=protected-access



from dagster import check
from dagster.core.instance import DagsterInstance
from dagster.serdes import ConfigurableClass

from .base import RunLauncher
from .cli_api_run_launcher import CliApiRunLauncher
from .grpc_run_launcher import GRPC_REPOSITORY_LOCATION_HANDLE_TYPES, GrpcRunLauncher


class DefaultRunLauncher(RunLauncher, ConfigurableClass):
    '''Default run launcher.

    This run launcher is aware of instance- and repository-level settings governing whether
    repositories should be loaded and runs launched over the legacy CLI API or over GRPC.
    '''

    def __init__(self, inst_data=None):
        self._inst_data = inst_data
        self._cli_api_run_launcher = CliApiRunLauncher()
        self._grpc_run_launcher = GrpcRunLauncher()

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data, _config_value):
        return DefaultRunLauncher(inst_data=inst_data,)

    def initialize(self, instance):
        check.inst_param(instance, 'instance', DagsterInstance)

        self._cli_api_run_launcher.initialize(instance)
        self._grpc_run_launcher.initialize(instance)

    def launch_run(self, instance, run, external_pipeline):
        repository_location_handle = external_pipeline.repository_handle.repository_location_handle
        if isinstance(repository_location_handle, GRPC_REPOSITORY_LOCATION_HANDLE_TYPES,):
            return self._grpc_run_launcher.launch_run(instance, run, external_pipeline)
        else:
            return self._cli_api_run_launcher.launch_run(instance, run, external_pipeline)

    def can_terminate(self, run_id):
        return self._cli_api_run_launcher.can_terminate(
            run_id
        ) or self._grpc_run_launcher.can_terminate(run_id)

    def terminate(self, run_id):
        return self._cli_api_run_launcher.terminate(run_id) or self._grpc_run_launcher.terminate(
            run_id
        )

    def join(self):
        self._cli_api_run_launcher.join()
        self._grpc_run_launcher.join()



from six import string_types

from dagster import Field, Permissive, Selector, Shape
from dagster.config.field_utils import _compute_fields_hash


def _hash(fields):
    return _compute_fields_hash(fields, description=None)


def test_compute_fields_hash():
    assert isinstance(_hash({'some_int': Field(int)}), string_types)


def test_hash_diff():

    assert _hash({'some_int': Field(int)}) != _hash({'another_int': Field(int)})

    assert _hash({'same_name': Field(int)}) != _hash({'same_name': Field(str)})

    assert _hash({'same_name': Field(int)}) != _hash({'same_name': Field(int, is_required=False)})

    assert _hash({'same_name': Field(int)}) != _hash(
        {'same_name': Field(int, is_required=False, default_value=2)}
    )

    assert _hash({'same_name': Field(int, is_required=False)}) != _hash(
        {'same_name': Field(int, is_required=False, default_value=2)}
    )

    assert _hash({'same_name': Field(int)}) != _hash({'same_name': Field(int, description='desc')})


def test_construct_same_dicts():
    int_dict_1 = Shape(fields={'an_int': Field(int)})
    int_dict_2 = Shape(fields={'an_int': Field(int)})

    # assert identical object
    assert int_dict_1 is int_dict_2
    # assert equivalent key
    assert int_dict_1.key == int_dict_2.key


def test_field_order_irrelevant():
    int_dict_1 = Shape(fields={'an_int': Field(int), 'another_int': Field(int)})

    int_dict_2 = Shape(fields={'another_int': Field(int), 'an_int': Field(int)})

    # assert identical object
    assert int_dict_1 is int_dict_2
    # assert equivalent key
    assert int_dict_1.key == int_dict_2.key


def test_construct_different_dicts():
    int_dict = Shape(fields={'an_int': Field(int)})
    string_dict = Shape(fields={'a_string': Field(str)})

    assert int_dict is not string_dict
    assert int_dict.key != string_dict.key


def test_construct_permissive_dict_same_same():
    assert Permissive() is Permissive()


def test_construct_same_perm_dicts():
    int_perm_dict_1 = Permissive(fields={'an_int': Field(int)})
    int_perm_dict_2 = Permissive(fields={'an_int': Field(int)})

    # assert identical object
    assert int_perm_dict_1 is int_perm_dict_2
    # assert equivalent key
    assert int_perm_dict_1.key == int_perm_dict_2.key


def test_construct_different_perm_dicts():
    int_perm_dict = Permissive(fields={'an_int': Field(int)})
    string_perm_dict = Permissive(fields={'a_string': Field(str)})

    assert int_perm_dict is not string_perm_dict
    assert int_perm_dict.key != string_perm_dict.key


def test_construct_same_selectors():
    int_selector_1 = Selector(fields={'an_int': Field(int)})
    int_selector_2 = Selector(fields={'an_int': Field(int)})

    # assert identical object
    assert int_selector_1 is int_selector_2
    # assert equivalent key
    assert int_selector_1.key == int_selector_2.key


def test_construct_different_selectors():
    int_selector = Selector(fields={'an_int': Field(int)})
    string_selector = Selector(fields={'a_string': Field(str)})

    assert int_selector is not string_selector
    assert int_selector.key != string_selector.key


def test_kitchen_sink():
    big_dict_1 = Shape(
        {
            'field_one': Field(int, default_value=2, is_required=False),
            'field_two': Field(
                Shape(
                    {
                        'nested_field_one': Field(bool),
                        'nested_selector': Field(
                            Selector(
                                {
                                    'int_field_in_selector': Field(int),
                                    'permissive_dict_in_selector': Field(Permissive()),
                                    'permissive_dict_with_fields_in_selector': Field(
                                        Permissive({'string_field': Field(str)})
                                    ),
                                }
                            )
                        ),
                    }
                )
            ),
        }
    )

    big_dict_2 = Shape(
        {
            'field_one': Field(int, default_value=2, is_required=False),
            'field_two': Field(
                Shape(
                    fields={
                        'nested_field_one': Field(bool),
                        'nested_selector': Field(
                            Selector(
                                fields={
                                    'permissive_dict_in_selector': Field(Permissive()),
                                    'int_field_in_selector': Field(int),
                                    'permissive_dict_with_fields_in_selector': Field(
                                        Permissive(fields={'string_field': Field(str)})
                                    ),
                                }
                            )
                        ),
                    }
                )
            ),
        }
    )

    assert big_dict_1 is big_dict_2
    assert big_dict_1.key == big_dict_2.key

    # differs way down in tree
    big_dict_3 = Shape(
        {
            'field_one': Field(int, default_value=2, is_required=False),
            'field_two': Field(
                Shape(
                    fields={
                        'nested_field_one': Field(bool),
                        'nested_selector': Field(
                            Selector(
                                fields={
                                    'permissive_dict_in_selector': Field(Permissive()),
                                    'int_field_in_selector': Field(int),
                                    'permissive_dict_with_fields_in_selector': Field(
                                        Permissive(fields={'int_field': Field(int)})
                                    ),
                                }
                            )
                        ),
                    }
                )
            ),
        }
    )

    assert big_dict_1 is not big_dict_3
    assert big_dict_1.key != big_dict_3.key



from dagster import check
from dagster.config.field_utils import check_user_facing_opt_config_param
from dagster.core.definitions.config import is_callable_valid_config_arg
from dagster.core.definitions.config_mappable import IConfigMappable
from dagster.utils.backcompat import rename_warning


class LoggerDefinition(IConfigMappable):
    '''Core class for defining loggers.

    Loggers are pipeline-scoped logging handlers, which will be automatically invoked whenever
    solids in a pipeline log messages.

    Args:
        logger_fn (Callable[[InitLoggerContext], logging.Logger]): User-provided function to
            instantiate the logger. This logger will be automatically invoked whenever the methods
            on ``context.log`` are called from within solid compute logic.
        config_schema (Optional[ConfigSchema]): The schema for the config. Configuration data available in
            `init_context.logger_config`.
        description (Optional[str]): A human-readable description of this logger.
        _configured_config_mapping_fn: This argument is for internal use only. Users should not
            specify this field. To preconfigure a resource, use the :py:func:`configured` API.
    '''

    def __init__(
        self, logger_fn, config_schema=None, description=None, _configured_config_mapping_fn=None,
    ):
        self._logger_fn = check.callable_param(logger_fn, 'logger_fn')
        self._config_schema = check_user_facing_opt_config_param(config_schema, 'config_schema')
        self._description = check.opt_str_param(description, 'description')
        self.__configured_config_mapping_fn = check.opt_callable_param(
            _configured_config_mapping_fn, 'config_mapping_fn'
        )

    @property
    def logger_fn(self):
        return self._logger_fn

    @property
    def config_field(self):
        rename_warning('config_schema', 'config_field', '0.9.0')
        return self._config_schema

    @property
    def config_schema(self):
        return self._config_schema

    @property
    def description(self):
        return self._description

    @property
    def _configured_config_mapping_fn(self):
        return self.__configured_config_mapping_fn

    def configured(self, config_or_config_fn, config_schema=None, **kwargs):
        wrapped_config_mapping_fn = self._get_wrapped_config_mapping_fn(
            config_or_config_fn, config_schema
        )

        return LoggerDefinition(
            config_schema=config_schema,
            description=kwargs.get('description', self.description),
            logger_fn=self.logger_fn,
            _configured_config_mapping_fn=wrapped_config_mapping_fn,
        )


def logger(config_schema=None, description=None):
    '''Define a logger.

    The decorated function should accept an :py:class:`InitLoggerContext` and return an instance of
    :py:class:`python:logging.Logger`. This function will become the ``logger_fn`` of an underlying
    :py:class:`LoggerDefinition`.

    Args:
        config_schema (Optional[ConfigSchema]): The schema for the config. Configuration data available in
            `init_context.logger_config`.
        description (Optional[str]): A human-readable description of the logger.
    '''
    # This case is for when decorator is used bare, without arguments.
    # E.g. @logger versus @logger()
    if callable(config_schema) and not is_callable_valid_config_arg(config_schema):
        return LoggerDefinition(logger_fn=config_schema)

    def _wrap(logger_fn):
        return LoggerDefinition(
            logger_fn=logger_fn, config_schema=config_schema, description=description,
        )

    return _wrap



from setuptools import find_packages, setup


def long_description():
    return '''
## Dagster
Dagster is a system for building modern data applications.

Combining an elegant programming model and beautiful tools, Dagster allows infrastructure engineers,
data engineers, and data scientists to seamlessly collaborate to process and produce the trusted,
reliable data needed in today's world.
'''.strip()


def get_version():
    version = {}
    with open('dagster/version.py') as fp:
        exec(fp.read(), version)  # pylint: disable=W0122

    return version['__version__']


if __name__ == '__main__':
    setup(
        name='dagster',
        version=get_version(),
        author='Elementl',
        author_email='hello@elementl.com',
        license='Apache-2.0',
        description='Dagster is an opinionated programming model for data pipelines.',
        long_description=long_description(),
        long_description_content_type='text/markdown',
        url='https://github.com/dagster-io/dagster',
        classifiers=[
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
        ],
        packages=find_packages(exclude=['dagster_tests']),
        package_data={
            'dagster': [
                'dagster/core/storage/event_log/sqlite/alembic/*',
                'dagster/core/storage/runs/sqlite/alembic/*',
                'dagster/core/storage/schedules/sqlite/alembic/*',
                'dagster/grpc/protos/*',
            ]
        },
        include_package_data=True,
        install_requires=[
            # standard python 2/3 compatability things
            'enum34; python_version < "3.4"',
            'future',
            'funcsigs',
            'functools32; python_version<"3"',
            'contextlib2>=0.5.4',
            'pathlib2>=2.3.4; python_version<"3"',
            # cli
            'click>=5.0',
            'coloredlogs>=6.1',
            'PyYAML',
            # core (not explicitly expressed atm)
            'alembic>=1.2.1',
            'grpcio==1.30.0',  # pinned so that code generated by grpc will work
            'protobuf==3.12.4',  # pinned so that code generated by grpc will work
            'pyrsistent>=0.14.8',
            'python-dateutil',
            'requests',
            'rx<=1.6.1',  # 3.0 was a breaking change. No py2 compatability as well.
            'futures; python_version < "3"',
            'six',
            'tqdm',
            'sqlalchemy>=1.0',
            'typing; python_version<"3"',
            'backports.tempfile; python_version<"3"',
            'toposort>=1.0',
            'watchdog>=0.8.3',
            'psutil >= 1.0; platform_system=="Windows"',
            # https://github.com/mhammond/pywin32/issues/1439
            'pywin32 != 226; platform_system=="Windows"',
            'pytz',
        ],
        extras_require={'docker': ['docker'],},
        entry_points={'console_scripts': ['dagster = dagster.cli:main']},
    )



from dagster import DagsterInstance, execute_pipeline, pipeline, reconstructable, solid


@solid(tags={'dagster/priority': '-1'})
def low(_):
    pass


@solid
def none(_):
    pass


@solid(tags={'dagster/priority': '1'})
def high(_):
    pass


@pipeline
def priority_test():
    none()
    low()
    high()
    none()
    low()
    high()


def test_priorities():

    result = execute_pipeline(priority_test,)
    assert result.success
    assert [
        str(event.solid_handle) for event in result.step_event_list if event.is_step_success
    ] == ['high', 'high_2', 'none', 'none_2', 'low', 'low_2']


def test_priorities_mp():
    pipe = reconstructable(priority_test)
    result = execute_pipeline(
        pipe,
        {
            'execution': {'multiprocess': {'config': {'max_concurrent': 1}}},
            'storage': {'filesystem': {}},
        },
        instance=DagsterInstance.local_temp(),
    )
    assert result.success
    assert [
        str(event.solid_handle) for event in result.step_event_list if event.is_step_success
    ] == ['high', 'high_2', 'none', 'none_2', 'low', 'low_2']



import subprocess

from dagster.grpc.client import DagsterGrpcClient
from dagster.grpc.server import wait_for_grpc_server
from dagster.utils import file_relative_path, find_free_port


def test_ping():
    port = find_free_port()
    python_file = file_relative_path(__file__, 'grpc_repo.py')
    process = subprocess.Popen(
        ['dagster', 'api', 'grpc', '--port', str(port), '--python-file', python_file],
        stdout=subprocess.PIPE,
    )

    try:
        wait_for_grpc_server(process)
        assert DagsterGrpcClient(port=port).ping('foobar') == 'foobar'
    finally:
        process.terminate()


def test_streaming():
    port = find_free_port()
    python_file = file_relative_path(__file__, 'grpc_repo.py')
    process = subprocess.Popen(
        ['dagster', 'api', 'grpc', '--port', str(port), '--python-file', python_file],
        stdout=subprocess.PIPE,
    )

    try:
        wait_for_grpc_server(process)
        api_client = DagsterGrpcClient(port=port)
        results = [result for result in api_client.streaming_ping(sequence_length=10, echo='foo')]
        assert len(results) == 10
        for sequence_number, result in enumerate(results):
            assert result['sequence_number'] == sequence_number
            assert result['echo'] == 'foo'
    finally:
        process.terminate()



from dagster import check
from dagster.config.config_type import ConfigType, ConfigTypeKind
from dagster.core.decorator_utils import (
    split_function_parameters,
    validate_decorated_fn_positionals,
)
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.utils import ensure_gen, ensure_single_item
from dagster.utils.backcompat import canonicalize_backcompat_args, rename_warning


class DagsterTypeLoader(object):
    @property
    def schema_type(self):
        check.not_implemented(
            'Must override schema_type in {klass}'.format(klass=type(self).__name__)
        )

    def construct_from_config_value(self, _context, config_value):
        '''
        How to create a runtime value from config data.
        '''
        return config_value

    def required_resource_keys(self):
        return frozenset()


class InputHydrationConfig(DagsterTypeLoader):
    def __init__(self):
        rename_warning('DagsterTypeLoader', 'InputHydrationConfig', '0.10.0')
        super(InputHydrationConfig, self).__init__()


class DagsterTypeMaterializer(object):
    @property
    def schema_type(self):
        check.not_implemented(
            'Must override schema_type in {klass}'.format(klass=type(self).__name__)
        )

    def materialize_runtime_values(self, _context, _config_value, _runtime_value):
        '''
        How to materialize a runtime value given configuration.
        '''
        check.not_implemented('Must implement')

    def required_resource_keys(self):
        return frozenset()


class OutputMaterializationConfig(DagsterTypeMaterializer):
    def __init__(self):
        rename_warning('DagsterTypeMaterializer', 'OutputMaterializationConfig', '0.10.0')
        super(OutputMaterializationConfig, self).__init__()


class DagsterTypeLoaderFromDecorator(DagsterTypeLoader):
    def __init__(self, config_type, func, required_resource_keys):
        self._config_type = check.inst_param(config_type, 'config_type', ConfigType)
        self._func = check.callable_param(func, 'func')
        self._required_resource_keys = check.opt_set_param(
            required_resource_keys, 'required_resource_keys', of_type=str
        )

    @property
    def schema_type(self):
        return self._config_type

    def construct_from_config_value(self, context, config_value):
        return self._func(context, config_value)

    def required_resource_keys(self):
        return frozenset(self._required_resource_keys)


def _create_type_loader_for_decorator(config_type, func, required_resource_keys):
    return DagsterTypeLoaderFromDecorator(config_type, func, required_resource_keys)


def input_hydration_config(config_schema=None, required_resource_keys=None, config_cls=None):
    '''Deprecated in favor of dagster_type_loader'''
    rename_warning('dagster_type_loader', 'input_hydration_config', '0.10.0')
    config_schema = canonicalize_backcompat_args(
        config_schema, 'config_schema', config_cls, 'config_cls', '0.10.0',
    )
    return dagster_type_loader(config_schema, required_resource_keys)


def dagster_type_loader(config_schema, required_resource_keys=None):
    '''Create an dagster type loader that maps config data to a runtime value.

    The decorated function should take the execution context and parsed config value and return the
    appropriate runtime value.

    Args:
        config_schema (ConfigSchema): The schema for the config that's passed to the decorated
            function.

    Examples:

    .. code-block:: python

        @dagster_type_loader(Permissive())
        def load_dict(_context, value):
            return value
    '''
    from dagster.config.field import resolve_to_config_type

    config_type = resolve_to_config_type(config_schema)
    EXPECTED_POSITIONALS = ['context', '*']

    def wrapper(func):
        fn_positionals, _ = split_function_parameters(func, EXPECTED_POSITIONALS)
        missing_positional = validate_decorated_fn_positionals(fn_positionals, EXPECTED_POSITIONALS)
        if missing_positional:
            raise DagsterInvalidDefinitionError(
                "@dagster_type_loader '{solid_name}' decorated function does not have required positional "
                "parameter '{missing_param}'. Solid functions should only have keyword arguments "
                "that match input names and a first positional parameter named 'context'.".format(
                    solid_name=func.__name__, missing_param=missing_positional
                )
            )
        return _create_type_loader_for_decorator(config_type, func, required_resource_keys)

    return wrapper


def input_selector_schema(config_cls, required_resource_keys=None):
    '''
    Deprecated in favor of dagster_type_loader.

    A decorator for annotating a function that can take the selected properties
    from a ``config_value`` in to an instance of a custom type.

    Args:
        config_cls (Selector)
    '''
    rename_warning('dagster_type_loader', 'input_selector_schema', '0.10.0')
    from dagster.config.field import resolve_to_config_type

    config_type = resolve_to_config_type(config_cls)
    check.param_invariant(config_type.kind == ConfigTypeKind.SELECTOR, 'config_cls')

    def _wrap(func):
        def _selector(context, config_value):
            selector_key, selector_value = ensure_single_item(config_value)
            return func(context, selector_key, selector_value)

        return _create_type_loader_for_decorator(config_type, _selector, required_resource_keys)

    return _wrap


class DagsterTypeMaterializerForDecorator(DagsterTypeMaterializer):
    def __init__(self, config_type, func, required_resource_keys):
        self._config_type = check.inst_param(config_type, 'config_type', ConfigType)
        self._func = check.callable_param(func, 'func')
        self._required_resource_keys = check.opt_set_param(
            required_resource_keys, 'required_resource_keys', of_type=str
        )

    @property
    def schema_type(self):
        return self._config_type

    def materialize_runtime_values(self, context, config_value, runtime_value):
        return ensure_gen(self._func(context, config_value, runtime_value))

    def required_resource_keys(self):
        return frozenset(self._required_resource_keys)


def _create_output_materializer_for_decorator(config_type, func, required_resource_keys):
    return DagsterTypeMaterializerForDecorator(config_type, func, required_resource_keys)


def output_materialization_config(config_schema=None, required_resource_keys=None, config_cls=None):
    '''Deprecated in favor of dagster_type_materializer'''
    rename_warning('dagster_type_materializer', 'output_materialization_config', '0.10.0')
    config_schema = canonicalize_backcompat_args(
        config_schema, 'config_schema', config_cls, 'config_cls', '0.10.0',
    )
    return dagster_type_materializer(config_schema, required_resource_keys)


def dagster_type_materializer(config_schema, required_resource_keys=None):
    '''Create an output materialization hydration config that configurably materializes a runtime
    value.

    The decorated function should take the execution context, the parsed config value, and the
    runtime value and the parsed config data, should materialize the runtime value, and should
    return an appropriate :py:class:`AssetMaterialization`.

    Args:
        config_schema (Any): The type of the config data expected by the decorated function.

    Examples:

    .. code-block:: python

        # Takes a list of dicts such as might be read in using csv.DictReader, as well as a config
        value, and writes
        @dagster_type_materializer(Path)
        def materialize_df(_context, path, value):
            with open(path, 'w') as fd:
                writer = csv.DictWriter(fd, fieldnames=value[0].keys())
                writer.writeheader()
                writer.writerows(rowdicts=value)

            return AssetMaterialization.file(path)

    '''
    from dagster.config.field import resolve_to_config_type

    config_type = resolve_to_config_type(config_schema)
    return lambda func: _create_output_materializer_for_decorator(
        config_type, func, required_resource_keys
    )


def output_selector_schema(config_cls, required_resource_keys=None):
    '''
    Deprecated in favor of dagster_type_materializer.

    A decorator for a annotating a function that can take the selected properties
    of a ``config_value`` and an instance of a custom type and materialize it.

    Args:
        config_cls (Selector):
    '''
    rename_warning('dagster_type_materializer', 'output_selector_schema', '0.10.0')
    from dagster.config.field import resolve_to_config_type

    config_type = resolve_to_config_type(config_cls)
    check.param_invariant(config_type.kind == ConfigTypeKind.SELECTOR, 'config_cls')

    def _wrap(func):
        def _selector(context, config_value, runtime_value):
            selector_key, selector_value = ensure_single_item(config_value)
            return func(context, selector_key, selector_value, runtime_value)

        return _create_output_materializer_for_decorator(
            config_type, _selector, required_resource_keys
        )

    return _wrap



from dagster import check
from dagster.config import Field
from dagster.core.definitions.intermediate_storage import intermediate_storage
from dagster.core.definitions.system_storage import SystemStorageData, system_storage
from dagster.core.storage.type_storage import TypeStoragePluginRegistry

from .file_manager import LocalFileManager
from .init import InitIntermediateStorageContext
from .intermediate_store import IntermediateStore, build_fs_intermediate_store
from .intermediates_manager import (
    IntermediateStoreIntermediatesManager,
    build_in_mem_intermediates_storage,
)
from .object_store import FilesystemObjectStore, InMemoryObjectStore, ObjectStore


def create_mem_system_storage_data(init_context):
    return SystemStorageData(
        intermediates_manager=build_in_mem_intermediates_storage(
            init_context.pipeline_run.run_id,
            type_storage_plugin_registry=init_context.type_storage_plugin_registry,
        ),
        file_manager=LocalFileManager.for_instance(
            init_context.instance, init_context.pipeline_run.run_id
        ),
    )


def build_intermediate_storage_from_object_store(
    object_store, init_context, root_for_run_id=lambda _: '',
):
    '''constructs an IntermediateStorage object from an object store and an init_context
        Call from within an intermediate_storage_definition
        Args:
            object_store(ObjectStore): The object store on which to base the intermediate store.
            init_context(InitIntermediateStorageContext):  the context from which to create the intermediates manager
            root_for_run_id_creator(Callable[[str], str]):
                        a function that converts from your run ID to the root of your object storage paths
        '''
    object_store = check.inst_param(object_store, 'object_store', ObjectStore)
    root_for_run_id = check.callable_param(root_for_run_id, 'root_for_run_id')
    init_context = check.inst_param(init_context, 'init_context', InitIntermediateStorageContext)

    intermediate_store = IntermediateStore(
        object_store=object_store,
        run_id=init_context.pipeline_run.run_id,
        root_for_run_id=root_for_run_id,
        type_storage_plugin_registry=init_context.type_storage_plugin_registry
        if init_context.type_storage_plugin_registry
        else TypeStoragePluginRegistry(types_to_register=[]),
    )
    return IntermediateStoreIntermediatesManager(intermediate_store)


@system_storage(name='in_memory', is_persistent=False, required_resource_keys=set())
def mem_system_storage(init_context):
    '''The default in-memory system storage.

    In most Dagster environments, this will be the default system storage. It is available by
    default on any :py:class:`ModeDefinition` that does not provide custom system storages. To
    select it explicitly, include the following top-level fragment in config:

    .. code-block:: yaml

        storage:
          in_memory:
    '''
    return create_mem_system_storage_data(init_context)


@intermediate_storage(name='in_memory', is_persistent=False, required_resource_keys=set())
def mem_intermediate_storage(init_context):
    '''The default in-memory intermediate storage.

    In-memory intermediate storage is the default on any pipeline run that does
    not configure any custom intermediate storage.

    Keep in mind when using this storage that intermediates will not be persisted after the pipeline
    run ends. Use a persistent intermediate storage like :py:func:`fs_intermediate_storage` to
    persist intermediates and take advantage of advanced features like pipeline re-execution.
    '''
    object_store = InMemoryObjectStore()
    return build_intermediate_storage_from_object_store(
        object_store=object_store, init_context=init_context
    )


@system_storage(
    name='filesystem',
    is_persistent=True,
    config_schema={'base_dir': Field(str, is_required=False)},
    required_resource_keys=set(),
)
def fs_system_storage(init_context):
    '''The default filesystem system storage.

    Filesystem system storage is available by default on any :py:class:`ModeDefinition` that does
    not provide custom system storages. To select it, include a fragment such as the following in
    config:

    .. code-block:: yaml

        storage:
          filesystem:
            base_dir: '/path/to/dir/'

    You may omit the ``base_dir`` config value, in which case the filesystem storage will use
    the :py:class:`DagsterInstance`-provided default.
    '''
    override_dir = init_context.system_storage_config.get('base_dir')
    if override_dir:
        file_manager = LocalFileManager(override_dir)
        intermediate_store = build_fs_intermediate_store(
            root_for_run_id=lambda _: override_dir,
            run_id=init_context.pipeline_run.run_id,
            type_storage_plugin_registry=init_context.type_storage_plugin_registry,
        )
    else:
        file_manager = LocalFileManager.for_instance(
            init_context.instance, init_context.pipeline_run.run_id
        )
        intermediate_store = build_fs_intermediate_store(
            init_context.instance.intermediates_directory,
            run_id=init_context.pipeline_run.run_id,
            type_storage_plugin_registry=init_context.type_storage_plugin_registry,
        )

    return SystemStorageData(
        file_manager=file_manager,
        intermediates_manager=IntermediateStoreIntermediatesManager(intermediate_store),
    )


@intermediate_storage(
    name='filesystem',
    is_persistent=True,
    config_schema={'base_dir': Field(str, is_required=False)},
    required_resource_keys=set(),
)
def fs_intermediate_storage(init_context):
    '''The default filesystem intermediate storage.

    Filesystem system storage is available by default on any :py:class:`ModeDefinition` that does
    not provide custom system storages. To select it, include a fragment such as the following in
    config:

    .. code-block:: yaml

        intermediate_storage:
          filesystem:
            base_dir: '/path/to/dir/'

    You may omit the ``base_dir`` config value, in which case the filesystem storage will use
    the :py:class:`DagsterInstance`-provided default.
    '''
    object_store = FilesystemObjectStore()
    override_dir = init_context.intermediate_storage_config.get('base_dir')
    if override_dir:
        root_for_run_id = lambda _: override_dir
    else:
        root_for_run_id = init_context.instance.intermediates_directory

    return build_intermediate_storage_from_object_store(
        object_store, init_context, root_for_run_id=root_for_run_id
    )


default_system_storage_defs = [mem_system_storage, fs_system_storage]

'''The default 'in_memory' and 'filesystem' system storage definitions.

Framework authors seeking to add their own system storage definitions can extend this list as follows:

.. code-block:: python

    custom_storage_mode = ModeDefinition(
        ...,
        system_storage_defs=default_system_storage_defs + [custom_system_storage_def]
    )
'''

default_intermediate_storage_defs = [mem_intermediate_storage, fs_intermediate_storage]
'''The default 'in_memory' and 'filesystem' intermediate storage definitions.

Framework authors seeking to add their own intermediate storage definitions can extend this list as follows:

.. code-block:: python

    custom_storage_mode = ModeDefinition(
        ...,
        intermediate_storage_defs=default_intermediate_storage_defs + [custom_intermediate_storage_def]
    )
'''



import uuid

import pytest

from dagster import (
    Field,
    ModeDefinition,
    RetryRequested,
    execute_pipeline,
    pipeline,
    reconstructable,
    resource,
    seven,
    solid,
)
from dagster.core.definitions.no_step_launcher import no_step_launcher
from dagster.core.events import DagsterEventType
from dagster.core.execution.api import create_execution_plan
from dagster.core.execution.context_creation_pipeline import PipelineExecutionContextManager
from dagster.core.execution.plan.external_step import (
    LocalExternalStepLauncher,
    local_external_step_launcher,
    step_context_to_step_run_ref,
    step_run_ref_to_step_context,
)
from dagster.core.execution.retries import Retries, RetryMode
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.utils.merger import deep_merge_dicts

RUN_CONFIG_BASE = {'solids': {'return_two': {'config': {'a': 'b'}}}}


def make_run_config(scratch_dir, mode):
    if mode in ['external', 'request_retry']:
        step_launcher_resource_keys = ['first_step_launcher', 'second_step_launcher']
    else:
        step_launcher_resource_keys = ['second_step_launcher']
    return deep_merge_dicts(
        RUN_CONFIG_BASE,
        {
            'resources': {
                step_launcher_resource_key: {'config': {'scratch_dir': scratch_dir}}
                for step_launcher_resource_key in step_launcher_resource_keys
            },
            'storage': {'filesystem': {'config': {'base_dir': scratch_dir}}},
        },
    )


class RequestRetryLocalExternalStepLauncher(LocalExternalStepLauncher):
    def launch_step(self, step_context, prior_attempts_count):
        if prior_attempts_count == 0:
            raise RetryRequested()
        else:
            return super(RequestRetryLocalExternalStepLauncher, self).launch_step(
                step_context, prior_attempts_count
            )


@resource(config_schema=local_external_step_launcher.config_schema)
def request_retry_local_external_step_launcher(context):
    return RequestRetryLocalExternalStepLauncher(**context.resource_config)


def define_basic_pipeline():
    @solid(required_resource_keys=set(['first_step_launcher']), config_schema={'a': Field(str)})
    def return_two(_):
        return 2

    @solid(required_resource_keys=set(['second_step_launcher']))
    def add_one(_, num):
        return num + 1

    @pipeline(
        mode_defs=[
            ModeDefinition(
                'external',
                resource_defs={
                    'first_step_launcher': local_external_step_launcher,
                    'second_step_launcher': local_external_step_launcher,
                },
            ),
            ModeDefinition(
                'internal_and_external',
                resource_defs={
                    'first_step_launcher': no_step_launcher,
                    'second_step_launcher': local_external_step_launcher,
                },
            ),
            ModeDefinition(
                'request_retry',
                resource_defs={
                    'first_step_launcher': request_retry_local_external_step_launcher,
                    'second_step_launcher': request_retry_local_external_step_launcher,
                },
            ),
        ]
    )
    def basic_pipeline():
        add_one(return_two())

    return basic_pipeline


def initialize_step_context(scratch_dir):
    pipeline_run = PipelineRun(
        pipeline_name='foo_pipeline',
        run_id=str(uuid.uuid4()),
        run_config=make_run_config(scratch_dir, 'external'),
        mode='external',
    )

    plan = create_execution_plan(
        reconstructable(define_basic_pipeline), pipeline_run.run_config, mode='external'
    )

    initialization_manager = PipelineExecutionContextManager(
        plan, pipeline_run.run_config, pipeline_run, DagsterInstance.ephemeral(),
    )
    for _ in initialization_manager.prepare_context():
        pass
    pipeline_context = initialization_manager.get_context()

    active_execution = plan.start(retries=Retries(RetryMode.DISABLED))
    step = active_execution.get_next_step()
    step_context = pipeline_context.for_step(step)
    return step_context


def test_step_context_to_step_run_ref():
    step_context = initialize_step_context('')
    step = step_context.step
    step_run_ref = step_context_to_step_run_ref(step_context, 0)
    assert step_run_ref.run_config == step_context.pipeline_run.run_config
    assert step_run_ref.run_id == step_context.pipeline_run.run_id

    rehydrated_step_context = step_run_ref_to_step_context(step_run_ref)
    assert rehydrated_step_context.required_resource_keys == step_context.required_resource_keys
    rehydrated_step = rehydrated_step_context.step
    assert rehydrated_step.pipeline_name == step.pipeline_name
    assert rehydrated_step.key_suffix == step.key_suffix
    assert rehydrated_step.step_inputs == step.step_inputs
    assert rehydrated_step.step_outputs == step.step_outputs
    assert rehydrated_step.kind == step.kind
    assert rehydrated_step.solid_handle.name == step.solid_handle.name
    assert rehydrated_step.logging_tags == step.logging_tags
    assert rehydrated_step.tags == step.tags


def test_local_external_step_launcher():
    with seven.TemporaryDirectory() as tmpdir:
        step_context = initialize_step_context(tmpdir)

        step_launcher = LocalExternalStepLauncher(tmpdir)
        events = list(step_launcher.launch_step(step_context, 0))
        event_types = [event.event_type for event in events]
        assert DagsterEventType.STEP_START in event_types
        assert DagsterEventType.STEP_SUCCESS in event_types
        assert DagsterEventType.STEP_FAILURE not in event_types


@pytest.mark.parametrize('mode', ['external', 'internal_and_external'])
def test_pipeline(mode):
    with seven.TemporaryDirectory() as tmpdir:
        result = execute_pipeline(
            pipeline=reconstructable(define_basic_pipeline),
            mode=mode,
            run_config=make_run_config(tmpdir, mode),
        )
        assert result.result_for_solid('return_two').output_value() == 2
        assert result.result_for_solid('add_one').output_value() == 3


def test_launcher_requests_retry():
    mode = 'request_retry'
    with seven.TemporaryDirectory() as tmpdir:
        result = execute_pipeline(
            pipeline=reconstructable(define_basic_pipeline),
            mode=mode,
            run_config=make_run_config(tmpdir, mode),
        )
        assert not result.success
        assert result.result_for_solid('return_two').output_value() == 2
        assert result.result_for_solid('add_one').output_value() == 3
        for step_key, events in result.events_by_step_key.items():
            if step_key:
                event_types = [event.event_type for event in events]
                assert DagsterEventType.STEP_UP_FOR_RETRY in event_types
                assert DagsterEventType.STEP_RESTARTED in event_types



import logging
from abc import abstractmethod
from collections import defaultdict

import six
import sqlalchemy as db

from dagster import check, seven
from dagster.core.definitions.events import AssetKey, Materialization
from dagster.core.errors import DagsterEventLogInvalidForRun
from dagster.core.events import DagsterEventType
from dagster.core.events.log import EventRecord
from dagster.core.execution.stats import RunStepKeyStatsSnapshot, StepEventStatus
from dagster.serdes import deserialize_json_to_dagster_namedtuple, serialize_dagster_namedtuple
from dagster.utils import datetime_as_float, utc_datetime_from_timestamp

from ..pipeline_run import PipelineRunStatsSnapshot
from .base import AssetAwareEventLogStorage, EventLogStorage
from .schema import SqlEventLogStorageTable


class SqlEventLogStorage(EventLogStorage):
    '''Base class for SQL backed event log storages.
    '''

    @abstractmethod
    def connect(self, run_id=None):
        '''Context manager yielding a connection.

        Args:
            run_id (Optional[str]): Enables those storages which shard based on run_id, e.g.,
                SqliteEventLogStorage, to connect appropriately.
        '''

    @abstractmethod
    def upgrade(self):
        '''This method should perform any schema or data migrations necessary to bring an
        out-of-date instance of the storage up to date.
        '''

    def prepare_insert_statement(self, event):
        ''' Helper method for preparing the event log SQL insertion statement.  Abstracted away to
        have a single place for the logical table representation of the event, while having a way
        for SQL backends to implement different execution implementations for `store_event`. See
        the `dagster-postgres` implementation which overrides the generic SQL implementation of
        `store_event`.
        '''

        dagster_event_type = None
        asset_key_str = None
        step_key = event.step_key

        if event.is_dagster_event:
            dagster_event_type = event.dagster_event.event_type_value
            step_key = event.dagster_event.step_key
            if event.dagster_event.asset_key:
                check.inst_param(event.dagster_event.asset_key, 'asset_key', AssetKey)
                asset_key_str = event.dagster_event.asset_key.to_string()

        # https://stackoverflow.com/a/54386260/324449
        return SqlEventLogStorageTable.insert().values(  # pylint: disable=no-value-for-parameter
            run_id=event.run_id,
            event=serialize_dagster_namedtuple(event),
            dagster_event_type=dagster_event_type,
            timestamp=utc_datetime_from_timestamp(event.timestamp),
            step_key=step_key,
            asset_key=asset_key_str,
        )

    def store_event(self, event):
        '''Store an event corresponding to a pipeline run.

        Args:
            event (EventRecord): The event to store.
        '''
        check.inst_param(event, 'event', EventRecord)
        sql_statement = self.prepare_insert_statement(event)
        run_id = event.run_id

        with self.connect(run_id) as conn:
            conn.execute(sql_statement)

    def get_logs_for_run_by_log_id(self, run_id, cursor=-1):
        check.str_param(run_id, 'run_id')
        check.int_param(cursor, 'cursor')
        check.invariant(
            cursor >= -1,
            'Don\'t know what to do with negative cursor {cursor}'.format(cursor=cursor),
        )

        # cursor starts at 0 & auto-increment column starts at 1 so adjust
        cursor = cursor + 1

        query = (
            db.select([SqlEventLogStorageTable.c.id, SqlEventLogStorageTable.c.event])
            .where(SqlEventLogStorageTable.c.run_id == run_id)
            .where(SqlEventLogStorageTable.c.id > cursor)
            .order_by(SqlEventLogStorageTable.c.id.asc())
        )

        with self.connect(run_id) as conn:
            results = conn.execute(query).fetchall()

        events = {}
        try:
            for (record_id, json_str,) in results:
                events[record_id] = check.inst_param(
                    deserialize_json_to_dagster_namedtuple(json_str), 'event', EventRecord
                )
        except (seven.JSONDecodeError, check.CheckError) as err:
            six.raise_from(DagsterEventLogInvalidForRun(run_id=run_id), err)

        return events

    def get_logs_for_run(self, run_id, cursor=-1):
        '''Get all of the logs corresponding to a run.

        Args:
            run_id (str): The id of the run for which to fetch logs.
            cursor (Optional[int]): Zero-indexed logs will be returned starting from cursor + 1,
                i.e., if cursor is -1, all logs will be returned. (default: -1)
        '''
        check.str_param(run_id, 'run_id')
        check.int_param(cursor, 'cursor')
        check.invariant(
            cursor >= -1,
            'Don\'t know what to do with negative cursor {cursor}'.format(cursor=cursor),
        )

        events_by_id = self.get_logs_for_run_by_log_id(run_id, cursor)
        return [event for id, event in sorted(events_by_id.items(), key=lambda x: x[0])]

    def get_stats_for_run(self, run_id):
        check.str_param(run_id, 'run_id')

        query = (
            db.select(
                [
                    SqlEventLogStorageTable.c.dagster_event_type,
                    db.func.count().label('n_events_of_type'),
                    db.func.max(SqlEventLogStorageTable.c.timestamp).label('last_event_timestamp'),
                ]
            )
            .where(SqlEventLogStorageTable.c.run_id == run_id)
            .group_by('dagster_event_type')
        )

        with self.connect(run_id) as conn:
            results = conn.execute(query).fetchall()

        try:
            counts = {}
            times = {}
            for result in results:
                (dagster_event_type, n_events_of_type, last_event_timestamp) = result
                if dagster_event_type:
                    counts[dagster_event_type] = n_events_of_type
                    times[dagster_event_type] = last_event_timestamp

            start_time = times.get(DagsterEventType.PIPELINE_START.value, None)
            end_time = times.get(
                DagsterEventType.PIPELINE_SUCCESS.value,
                times.get(DagsterEventType.PIPELINE_FAILURE.value, None),
            )

            return PipelineRunStatsSnapshot(
                run_id=run_id,
                steps_succeeded=counts.get(DagsterEventType.STEP_SUCCESS.value, 0),
                steps_failed=counts.get(DagsterEventType.STEP_FAILURE.value, 0),
                materializations=counts.get(DagsterEventType.STEP_MATERIALIZATION.value, 0),
                expectations=counts.get(DagsterEventType.STEP_EXPECTATION_RESULT.value, 0),
                start_time=datetime_as_float(start_time) if start_time else None,
                end_time=datetime_as_float(end_time) if end_time else None,
            )
        except (seven.JSONDecodeError, check.CheckError) as err:
            six.raise_from(DagsterEventLogInvalidForRun(run_id=run_id), err)

    def get_step_stats_for_run(self, run_id):
        check.str_param(run_id, 'run_id')

        STEP_STATS_EVENT_TYPES = [
            DagsterEventType.STEP_START.value,
            DagsterEventType.STEP_SUCCESS.value,
            DagsterEventType.STEP_SKIPPED.value,
            DagsterEventType.STEP_FAILURE.value,
        ]

        by_step_query = (
            db.select(
                [
                    SqlEventLogStorageTable.c.step_key,
                    SqlEventLogStorageTable.c.dagster_event_type,
                    db.func.max(SqlEventLogStorageTable.c.timestamp).label('timestamp'),
                ]
            )
            .where(SqlEventLogStorageTable.c.run_id == run_id)
            .where(SqlEventLogStorageTable.c.step_key != None)
            .where(SqlEventLogStorageTable.c.dagster_event_type.in_(STEP_STATS_EVENT_TYPES))
            .group_by(
                SqlEventLogStorageTable.c.step_key, SqlEventLogStorageTable.c.dagster_event_type,
            )
        )

        with self.connect(run_id) as conn:
            results = conn.execute(by_step_query).fetchall()

        by_step_key = defaultdict(dict)
        for result in results:
            step_key = result.step_key
            if result.dagster_event_type == DagsterEventType.STEP_START.value:
                by_step_key[step_key]['start_time'] = (
                    datetime_as_float(result.timestamp) if result.timestamp else None
                )
            if result.dagster_event_type == DagsterEventType.STEP_FAILURE.value:
                by_step_key[step_key]['end_time'] = (
                    datetime_as_float(result.timestamp) if result.timestamp else None
                )
                by_step_key[step_key]['status'] = StepEventStatus.FAILURE
            if result.dagster_event_type == DagsterEventType.STEP_SUCCESS.value:
                by_step_key[step_key]['end_time'] = (
                    datetime_as_float(result.timestamp) if result.timestamp else None
                )
                by_step_key[step_key]['status'] = StepEventStatus.SUCCESS
            if result.dagster_event_type == DagsterEventType.STEP_SKIPPED.value:
                by_step_key[step_key]['end_time'] = (
                    datetime_as_float(result.timestamp) if result.timestamp else None
                )
                by_step_key[step_key]['status'] = StepEventStatus.SKIPPED

        materializations = defaultdict(list)
        expectation_results = defaultdict(list)
        raw_event_query = (
            db.select([SqlEventLogStorageTable.c.event])
            .where(SqlEventLogStorageTable.c.run_id == run_id)
            .where(SqlEventLogStorageTable.c.step_key != None)
            .where(
                SqlEventLogStorageTable.c.dagster_event_type.in_(
                    [
                        DagsterEventType.STEP_MATERIALIZATION.value,
                        DagsterEventType.STEP_EXPECTATION_RESULT.value,
                    ]
                )
            )
            .order_by(SqlEventLogStorageTable.c.id.asc())
        )

        with self.connect(run_id) as conn:
            results = conn.execute(raw_event_query).fetchall()

        try:
            for (json_str,) in results:
                event = check.inst_param(
                    deserialize_json_to_dagster_namedtuple(json_str), 'event', EventRecord
                )
                if event.dagster_event.event_type == DagsterEventType.STEP_MATERIALIZATION:
                    materializations[event.step_key].append(
                        event.dagster_event.event_specific_data.materialization
                    )
                elif event.dagster_event.event_type == DagsterEventType.STEP_EXPECTATION_RESULT:
                    expectation_results[event.step_key].append(
                        event.dagster_event.event_specific_data.expectation_result
                    )
        except (seven.JSONDecodeError, check.CheckError) as err:
            six.raise_from(DagsterEventLogInvalidForRun(run_id=run_id), err)

        return [
            RunStepKeyStatsSnapshot(
                run_id=run_id,
                step_key=step_key,
                status=value.get('status'),
                start_time=value.get('start_time'),
                end_time=value.get('end_time'),
                materializations=materializations.get(step_key),
                expectation_results=expectation_results.get(step_key),
            )
            for step_key, value in by_step_key.items()
        ]

    def wipe(self):
        '''Clears the event log storage.'''
        # Should be overridden by SqliteEventLogStorage and other storages that shard based on
        # run_id
        # https://stackoverflow.com/a/54386260/324449
        with self.connect() as conn:
            conn.execute(SqlEventLogStorageTable.delete())  # pylint: disable=no-value-for-parameter

    def delete_events(self, run_id):
        check.str_param(run_id, 'run_id')

        statement = SqlEventLogStorageTable.delete().where(  # pylint: disable=no-value-for-parameter
            SqlEventLogStorageTable.c.run_id == run_id
        )

        with self.connect(run_id) as conn:
            conn.execute(statement)

    @property
    def is_persistent(self):
        return True

    def update_event_log_record(self, record_id, event):
        ''' Utility method for migration scripts to update SQL representation of event records. '''
        check.int_param(record_id, 'record_id')
        check.inst_param(event, 'event', EventRecord)
        dagster_event_type = None
        asset_key_str = None
        if event.is_dagster_event:
            dagster_event_type = event.dagster_event.event_type_value
            if event.dagster_event.asset_key:
                check.inst_param(event.dagster_event.asset_key, 'asset_key', AssetKey)
                asset_key_str = event.dagster_event.asset_key.to_string()

        with self.connect(run_id=event.run_id) as conn:
            conn.execute(
                SqlEventLogStorageTable.update()  # pylint: disable=no-value-for-parameter
                .where(SqlEventLogStorageTable.c.id == record_id)
                .values(
                    event=serialize_dagster_namedtuple(event),
                    dagster_event_type=dagster_event_type,
                    timestamp=utc_datetime_from_timestamp(event.timestamp),
                    step_key=event.step_key,
                    asset_key=asset_key_str,
                )
            )

    def get_event_log_table_data(self, run_id, record_id):
        ''' Utility method to test representation of the record in the SQL table.  Returns all of
        the columns stored in the event log storage (as opposed to the deserialized `EventRecord`).
        This allows checking that certain fields are extracted to support performant lookups (e.g.
        extracting `step_key` for fast filtering)'''
        with self.connect(run_id=run_id) as conn:
            query = (
                db.select([SqlEventLogStorageTable])
                .where(SqlEventLogStorageTable.c.id == record_id)
                .order_by(SqlEventLogStorageTable.c.id.asc())
            )
            return conn.execute(query).fetchone()


class AssetAwareSqlEventLogStorage(AssetAwareEventLogStorage, SqlEventLogStorage):
    @abstractmethod
    def connect(self, run_id=None):
        pass

    @abstractmethod
    def upgrade(self):
        pass

    def _add_cursor_limit_to_query(self, query, cursor, limit):
        ''' Helper function to deal with cursor/limit pagination args '''

        if cursor:
            cursor_query = db.select([SqlEventLogStorageTable.c.id]).where(
                SqlEventLogStorageTable.c.id == cursor
            )
            query = query.where(SqlEventLogStorageTable.c.id < cursor_query)

        if limit:
            query = query.limit(limit)

        query = query.order_by(SqlEventLogStorageTable.c.timestamp.desc())
        return query

    def get_all_asset_keys(self):
        query = db.select([SqlEventLogStorageTable.c.asset_key]).distinct()
        with self.connect() as conn:
            results = conn.execute(query).fetchall()

        return [AssetKey.from_db_string(asset_key) for (asset_key,) in results if asset_key]

    def get_asset_events(self, asset_key, cursor=None, limit=None):
        check.inst_param(asset_key, 'asset_key', AssetKey)
        asset_key_str = asset_key.to_string()
        query = db.select([SqlEventLogStorageTable.c.id, SqlEventLogStorageTable.c.event]).where(
            SqlEventLogStorageTable.c.asset_key == asset_key_str
        )
        query = self._add_cursor_limit_to_query(query, cursor, limit)
        with self.connect() as conn:
            results = conn.execute(query).fetchall()

        events = []
        for row_id, json_str in results:
            try:
                event_record = deserialize_json_to_dagster_namedtuple(json_str)
                if not isinstance(event_record, EventRecord):
                    logging.warning(
                        'Could not resolve asset event record as EventRecord for id `{}`.'.format(
                            row_id
                        )
                    )
                    continue
                events.append(event_record)
            except seven.JSONDecodeError:
                logging.warning('Could not parse asset event record id `{}`.'.format(row_id))
        return events

    def get_asset_run_ids(self, asset_key):
        check.inst_param(asset_key, 'asset_key', AssetKey)
        asset_key_str = asset_key.to_string()
        query = (
            db.select(
                [SqlEventLogStorageTable.c.run_id, db.func.max(SqlEventLogStorageTable.c.timestamp)]
            )
            .where(SqlEventLogStorageTable.c.asset_key == asset_key_str)
            .group_by(SqlEventLogStorageTable.c.run_id,)
            .order_by(db.func.max(SqlEventLogStorageTable.c.timestamp).desc())
        )

        with self.connect() as conn:
            results = conn.execute(query).fetchall()

        return [run_id for (run_id, _timestamp) in results]

    def wipe_asset(self, asset_key):
        check.inst_param(asset_key, 'asset_key', AssetKey)
        query = db.select([SqlEventLogStorageTable.c.id, SqlEventLogStorageTable.c.event]).where(
            SqlEventLogStorageTable.c.asset_key == asset_key.to_string()
        )
        with self.connect() as conn:
            results = conn.execute(query).fetchall()

        for row_id, json_str in results:
            try:
                event_record = deserialize_json_to_dagster_namedtuple(json_str)
                if not isinstance(event_record, EventRecord):
                    continue

                assert event_record.dagster_event.event_specific_data.materialization.asset_key

                dagster_event = event_record.dagster_event
                event_specific_data = dagster_event.event_specific_data
                materialization = event_specific_data.materialization
                updated_materialization = Materialization(
                    label=materialization.label,
                    description=materialization.description,
                    metadata_entries=materialization.metadata_entries,
                    asset_key=None,
                    skip_deprecation_warning=True,
                )
                updated_event_specific_data = event_specific_data._replace(
                    materialization=updated_materialization
                )
                updated_dagster_event = dagster_event._replace(
                    event_specific_data=updated_event_specific_data
                )
                updated_record = event_record._replace(dagster_event=updated_dagster_event)

                # update the event_record here
                self.update_event_log_record(row_id, updated_record)

            except seven.JSONDecodeError:
                logging.warning('Could not parse asset event record id `{}`.'.format(row_id))



from abc import ABCMeta

import six

from dagster import check
from dagster.core.execution.context.system import SystemExecutionContext
from dagster.core.types.dagster_type import DagsterType

from .object_store import FilesystemObjectStore, InMemoryObjectStore, ObjectStore
from .type_storage import TypeStoragePluginRegistry


class IntermediateStore(six.with_metaclass(ABCMeta)):
    def __init__(self, object_store, root_for_run_id, run_id, type_storage_plugin_registry):
        self.root_for_run_id = check.callable_param(root_for_run_id, 'root_for_run_id')
        self.run_id = check.str_param(run_id, 'run_id')
        self.object_store = check.inst_param(object_store, 'object_store', ObjectStore)
        self.type_storage_plugin_registry = check.inst_param(
            type_storage_plugin_registry, 'type_storage_plugin_registry', TypeStoragePluginRegistry
        )

    @property
    def root(self):
        return self.root_for_run_id(self.run_id)

    def uri_for_paths(self, paths, protocol=None):
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')
        key = self.key_for_paths(paths)
        return self.object_store.uri_for_key(key, protocol)

    def key_for_paths(self, paths):
        return self.object_store.key_for_paths([self.root] + paths)

    def set_object(self, obj, context, dagster_type, paths):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.inst_param(dagster_type, 'dagster_type', DagsterType)
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')
        key = self.object_store.key_for_paths([self.root] + paths)
        return self.object_store.set_object(
            key, obj, serialization_strategy=dagster_type.serialization_strategy
        )

    def get_object(self, context, dagster_type, paths):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')
        check.inst_param(dagster_type, 'dagster_type', DagsterType)
        key = self.object_store.key_for_paths([self.root] + paths)
        return self.object_store.get_object(
            key, serialization_strategy=dagster_type.serialization_strategy
        )

    def has_object(self, context, paths):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')
        key = self.object_store.key_for_paths([self.root] + paths)
        return self.object_store.has_object(key)

    def rm_object(self, context, paths):
        check.opt_inst_param(context, 'context', SystemExecutionContext)
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')
        key = self.object_store.key_for_paths([self.root] + paths)
        self.object_store.rm_object(key)

    def copy_object_from_run(self, _context, run_id, paths):
        check.str_param(run_id, 'run_id')
        check.list_param(paths, 'paths', of_type=str)
        check.param_invariant(len(paths) > 0, 'paths')

        src = self.object_store.key_for_paths([self.root_for_run_id(run_id)] + paths)
        dst = self.object_store.key_for_paths([self.root] + paths)

        return self.object_store.cp_object(src, dst)

    def set_value(self, obj, context, dagster_type, paths):
        if self.type_storage_plugin_registry.is_registered(dagster_type):
            return self.type_storage_plugin_registry.get(dagster_type.name).set_object(
                self, obj, context, dagster_type, paths
            )
        elif dagster_type.name is None:
            self.type_storage_plugin_registry.check_for_unsupported_composite_overrides(
                dagster_type
            )

        return self.set_object(obj, context, dagster_type, paths)

    def get_value(self, context, dagster_type, paths):
        if self.type_storage_plugin_registry.is_registered(dagster_type):
            return self.type_storage_plugin_registry.get(dagster_type.name).get_object(
                self, context, dagster_type, paths
            )
        elif dagster_type.name is None:
            self.type_storage_plugin_registry.check_for_unsupported_composite_overrides(
                dagster_type
            )
        return self.get_object(context, dagster_type, paths)


def build_fs_intermediate_store(root_for_run_id, run_id, type_storage_plugin_registry=None):
    return IntermediateStore(
        FilesystemObjectStore(),
        root_for_run_id,
        run_id,
        type_storage_plugin_registry
        if type_storage_plugin_registry
        else TypeStoragePluginRegistry(types_to_register=[]),
    )


def build_mem_intermediate_store(run_id, type_storage_plugin_registry=None):
    return IntermediateStore(
        InMemoryObjectStore(),
        lambda _: '',
        run_id,
        type_storage_plugin_registry
        if type_storage_plugin_registry
        else TypeStoragePluginRegistry(types_to_register=[]),
    )
